#include "Actor.h"

void drawActor(Actor* actor, u8* actorSprite)
{
    u8 cellX = actor->object.position.x / TILE_WIDTH_BYTES;
    u8 cellY = actor->object.position.y / TILE_HEIGHT;
    u8 previousCellX = actor->object.previousPosition.x / TILE_WIDTH_BYTES;
    u8 previousCellY = actor->object.previousPosition.y / TILE_HEIGHT;

    // Make sure to calculate the backbuffer width based on previous cell otherwise
    // old data can be displayed on the left side of the screen when it is first drawn
    // when coming into a room from the right.
    backbufferWidth = previousCellX == 19 ? TILE_WIDTH_BYTES : BACKBUFFER_WIDTH;

    // Populate the backbuffer from the previous actor position and draw on the screen.
    if (previousCellX != cellX || previousCellY != cellY)
    {
        populateBackbufferFromMap(previousCellX, previousCellY);
        drawBackbuffer(previousCellX, previousCellY);
    }

    // When at the right edge of the screen (x == 19) the width of the back buffer is only one cell wide.
    backbufferWidth = cellX == 19 ? TILE_WIDTH_BYTES : BACKBUFFER_WIDTH;

    // Add the sprites to the back buffer (we expand x and y to one of the 4 cells in the back buffer but the source is the map grid for the cell value).
    populateBackbufferFromMap(cellX, cellY);

    // Add the actor sprite to the back buffer.
    addSpriteToBackbuffer(actor->object.position.x % TILE_WIDTH_BYTES, actor->object.position.y % TILE_HEIGHT, actorSprite);

    // Draw the back buffer now that the actor sprite has been added to it.
    drawBackbuffer(cellX, cellY);

    // This is to prevent the player disappearing when the enemy backbuffer is drawn
    // as I do not have code to actually add the player to the backbuffer itself.
    if (actor->actorType == ACTOR_TYPE_ENEMY)
    {
        // Check if the player could be within the enemy backbuffer.
        if (difference(actor->object.position.y, player.object.position.y) < 32 && difference(actor->object.position.x, player.object.position.x) < 8)
        {
            // Draw the player without the backbuffer on screen.
            drawActorNoBackbuffer(&player.object, playerGraphics[player.graphicIndex]);
        }
    }
}

void drawActorNoBackbuffer(Object* object, u8* actorSprite)
{
    u8* memory = cpct_getScreenPtr(CPCT_VMEM_START, object->position.x, object->position.y);
    cpct_drawSprite(actorSprite, memory, TILE_WIDTH_BYTES, TILE_HEIGHT);
}

void activateActorLink(Actor* actor, u8 currentLinkX, u8 currentLinkY)
{
    // Determine whether the link should be activated or deactivated based on whether the actor is moving onto the plate or off it.
    if (currentLinkX == NO_LINK_FOUND)
    {
        if (actor->activatedLinkX != NO_LINK_FOUND)
        {
            // New link source requiring the new link to be activated.
            activateLinkedInteractable(actor->activatedLinkX, actor->activatedLinkY);
        }

        // If both are no link then no action required.
    }
    else
    {
        // Current link has been set.

        if (actor->activatedLinkX == NO_LINK_FOUND)
        {
            // Moved off a link requiring it to be deactivated.
            activateLinkedInteractable(currentLinkX, currentLinkY);
        }
        else if (actor->activatedLinkX != currentLinkX || actor->activatedLinkY != currentLinkY)
        {
            // Moved to a new link source requiring the old link to be deactivated and the new link to be activated.
            activateLinkedInteractable(currentLinkX, currentLinkY);
            activateLinkedInteractable(actor->activatedLinkX, actor->activatedLinkY);
        }

        // If both are the same then no action.
    }
}

u8 actorOccuppyingCells(Actor* actor, u8 cellType)
{
    u8 cellDetected = 0;

    // Find the cells that the actor is in.
    // Some of these may be duplicate but we are looking for when the actor occupies all four cells.
    u8 occupiedXTileLeft = (actor->object.position.x) >> 2;
    u8 occupiedXTileRight = (actor->object.position.x + TILE_WIDTH_BYTES - 1);
    u8 occupiedYTileBottom = (actor->object.position.y + TILE_HEIGHT - 1);
    u8 occupiedYTileTop = (actor->object.position.y) >> 4;
    occupiedXTileRight >>= 2;
    occupiedYTileBottom >>= 4;

    // Check that the actor occupies every cell.
    cellDetected += cellContainsType(occupiedXTileLeft, occupiedYTileTop, cellType);
    cellDetected += cellContainsType(occupiedXTileRight, occupiedYTileTop, cellType);
    cellDetected += cellContainsType(occupiedXTileLeft, occupiedYTileBottom, cellType);
    cellDetected += cellContainsType(occupiedXTileRight, occupiedYTileBottom, cellType);

    // This is expected to be 4 if the actor is completely occupying the specified cell type.
    return cellDetected;
}

u8 nextActorId()
{
    return currentId++;
}

void initActor(Actor* actor, u8 x, u8 y, u8 actorType)
{
    u8 roomX = x / SCREEN_TILES_X;
    u8 roomY = y / SCREEN_TILES_Y;

    // Set room and position information based on x,y values.
    actor->object.position.x = (x % SCREEN_TILES_X) * TILE_WIDTH_BYTES;
    actor->object.position.y = (y % SCREEN_TILES_Y) * TILE_HEIGHT;
    actor->object.previousPosition.x = actor->object.position.x;
    actor->object.previousPosition.y = actor->object.position.y;
    actor->object.position.room = roomX + roomY * 3;
    actor->object.previousPosition.room = actor->object.position.room;

    // For enemy and block activating a link for pressure plates.
    actor->activatedLinkX = NO_LINK_FOUND;

    // Set the unique id for the actor.
    actor->id = nextActorId();

    // Set redraw to draw the actor by default.
    actor->redraw = TRUE;

    // The type of actor being initialised.
    actor->actorType = actorType;
}