#ifndef ENEMY_H
#define ENEMY_H

#include <cpctelera.h>
#include "GameTypes.h"
#include "Interactable.h"
#include "Object.h"
#include "sprites/EnemySprites.h"

typedef struct Enemy Enemy;

struct Enemy
{
    // Actor object containing position and velocity.
    Object object;

    // The x position of the link activated by the enemy.
    u8 activatedLinkX;

    // The y position of the link activated by the enemy.
    u8 activatedLinkY;

    // Number of hitpoints.
    u8 health;

    // Unique identifier for identifying the different actors in the system.
    u8 id;

    // Indicates whether the actor should be redrawn on screen.
    u8 redraw;

    // The type of actor (player, enemy, block).
    u8 actorType;

    // Indicates whether the enemy collided or is near collision with other actors for drawing purposes.
    u8 collision;

    // The type of enemy.
    u8 type;

    // For reducing the frequency that the enemy logic runs thus improving performance (also allows for speed adjustments).
    u8 logicFlag;

    // The index of the projectile belonging to the enemy.
    u8 projectileIndex;

    // The direction that the enemy turret is currently facing.
    u8 direction;
};

// TODO real value, Maximum number of enemies that can be loaded from a map.
#define MAX_ENEMY_COUNT 15

// The number of components of the enemy array that populate the actual enemy object.
#define ENEMY_ARRAY_COMPONENT_COUNT 3

// An enemy that is moving up and down on the screen\level.
#define ENEMY_UP_DOWN 1

// An enemy that is moving left and right on the screen\level.
#define ENEMY_LEFT_RIGHT 2

// An enemy that turns the corner (ninety degrees) every time it is blocked.
#define ENEMY_CORNERING 3

// An enemy that moves leftwards the player in a straight line and crushes them on collision.
#define ENEMY_MOVING_ROCK_LEFT 4

// An enemy that moves rightwards the player in a straight line and crushes them on collision.
#define ENEMY_MOVING_ROCK_RIGHT 5

// An enemy that moves upwards the player in a straight line and crushes them on collision.
#define ENEMY_MOVING_ROCK_UP 6

// An enemy that moves downwards the player in a straight line and crushes them on collision.
#define ENEMY_MOVING_ROCK_DOWN 7

// An enemy that remains in the water and attacks the player when they enter the water.
#define ENEMY_SHARK 8

// A stationary turret that fires projectiles left.
#define ENEMY_TURRET_LEFT 9

// A stationary turret that fires projectiles right.
#define ENEMY_TURRET_RIGHT 10

// A stationary turret that fires projectiles upward.
#define ENEMY_TURRET_UP 11

// A stationary turret that fires projectiles downward.
#define ENEMY_TURRET_DOWN 12

// A stationary spike that goes up and down over a set interval.
#define ENEMY_TIMED_SPIKE 13

// The number of enemy entities on the map.
extern u8 mapEnemyCount;

// The enemy entities on the map.
extern Enemy mapEnemies[MAX_ENEMY_COUNT];

// Initialise the enemies to their positions on the map.
void initEnemies(u8* enemiesArray, u8 arraySize);

// Draw the enemies on the screen.
void drawEnemies();

// Marks the enemies in the enemies array for redraw.
void markEnemiesForRedraw();

// Run enemy logic, performing enemy actions and moving.
void logicEnemies();

// Update a specified enemy's position based on velocity and whether it is impeded (collision detection).
void updateEnemyPosition(Enemy* enemy);

// Processes the cell and indicates whether it is solid (can the enemy pass through).
u8 processCellForEnemy(u8 x, u8 y, Enemy* enemy);

// Converts an enemy to a timed spike.
void convertEnemyToSpike(Enemy* enemy);

// Destroys the enemy (removes them from screen).
void destroyEnemy(Enemy* enemy);

// Spawns a projectile from an enemy position.
u8 spawnEnemyProjectile(Enemy* enemy, u8 enemyIndex);

#endif