#include "Environment.h"

// The maximum number of elements in the crumble queue (player can occupy 4 squares).
#define QUEUE_SIZE 4

// Crumble queue item is empty.
#define EMPTY_QUEUE_ITEM 9999

// All items in the crumble queue are occupied.
#define QUEUE_FULL 255

// Crumble queue for holding the environment cells that are currently crumbling.
u16 crumbleQueue_[QUEUE_SIZE];

// The hedge cell has no value.
#define HEDGE_CELL_UNSET 9999

// The environment hedge cell being attacked with a machete.
u16 hedgeCell_ = HEDGE_CELL_UNSET;

void crumbleBlock(u8 x, u8 y, u8 roomX, u8 roomY)
{
    // The index in the map array that is crumbling.
    u16 mapCellIndex = mapCellPosition(x, y, roomX, roomY);

    // Attempts to queue the map cell onto the crumble queue as long as the queue is not full.
    if (queueItem(mapCellIndex) != QUEUE_FULL)
    {
        // Trigger the crumbling timer.
        crumblingTimerActive = TRUE;
    }
    else
    {
        // Queue is full so transition everything to the next crumbling state (each item transitioned will be cleared from the queue).
        transitionCrumblingBlock();

        // Add the new cell to the now cleared crumble queue.
        queueItem(mapCellIndex);
    }
}

void transitionCrumblingBlock()
{
    u8 i = 0;

    // Transition all the blocks in the queue.
    for (i = 0; i < QUEUE_SIZE; ++i)
    {
        // Empty queue items will be ignored.
        if (crumbleQueue_[i] != EMPTY_QUEUE_ITEM)
        {
            u16 mapCellsIndex = crumbleQueue_[i];
            u8 cellValue = mapCells[mapCellsIndex];
            u8 cellX = 0;
            u8 cellY = 0;

            // Transition block to the level crumble cell when it is at the end of the cycle.
            // Changed to >= instead of == as this got turned into turret cell instead of water on rare occasions.
            // The >= appeared to solve this problem.
            if (cellValue >= CELL_ENVIRONMENT_FLAG + ENVIRONMENT_CRUMBLING_BLOCK_1)
            {
                mapCells[mapCellsIndex] = endCrumbleCell;
            }
            else
            {
                // Increment the value to move to the next crumbling block type.
                ++mapCells[mapCellsIndex];
            }

            // Convert from the map cell index value to the x and y values in the current room.
            cellX = (mapCellsIndex / MAP_TILES_Y) % SCREEN_TILES_X;
            cellY = (mapCellsIndex % MAP_TILES_X) % SCREEN_TILES_Y;

            // Draw the updated map cell at the position.
            drawCellAtPosition(cellX, cellY);

            // Set queue item to empty (dequeue the item).
            crumbleQueue_[i] = EMPTY_QUEUE_ITEM;

            // Play a sound when the block crumbles to the next stage.
            playSoundEffect(SFX_CRUMBLE, CHANNEL_A, 4);
        }
    }

    // Redraw the player so that it is not covered up by the crumbling block.
    player.redraw = TRUE;
}

void triggerFlow(u8 x, u8 y, u8 roomX, u8 roomY, u8 sourceCell, u8 destinationCell)
{
    // Check that the current cell is lava flow.
    u8 cell = mapCell(x, y, roomX, roomY);

    if (cell == sourceCell)
    {
        // Update the source flow cell to destination cell value.
        updateMapCellRedraw(x, y, roomX, roomY, destinationCell);

        // Checks for source flow above the current cell and converts source flow cell to destination cell.
        if (y + 1 < SCREEN_TILES_Y)
        {
            triggerFlow(x, y + 1, roomX, roomY, sourceCell, destinationCell);
        }

        // Check for source flow below the current cell and converts source flow to destination cell.
        if (y - 1 >= 0)
        {
            triggerFlow(x, y - 1, roomX, roomY, sourceCell, destinationCell);
        }

        // Check for source flow right of the current cell and converts source flow to destination cell.
        if (x + 1 < SCREEN_TILES_X)
        {
            triggerFlow(x + 1, y, roomX, roomY, sourceCell, destinationCell);
        }

        // Check for source flow left of the current cell and converts source flow to destination cell.
        if (x - 1 >= 0)
        {
            triggerFlow(x - 1, y, roomX, roomY, sourceCell, destinationCell);
        }
    }
}

void initCrumbleQueue()
{
    u8 i = 0;

    for (i = 0; i < QUEUE_SIZE; ++i)
    {
        crumbleQueue_[i] = EMPTY_QUEUE_ITEM;
    }
}

u8 queueItem(u16 mapCellIndex)
{
    u8 i = 0;

    for (i = 0; i < QUEUE_SIZE; ++i)
    {
        if (crumbleQueue_[i] == EMPTY_QUEUE_ITEM)
        {
            // Queue index empty so add it to the queue.
            crumbleQueue_[i] = mapCellIndex;
            return i;
        }
        else if (crumbleQueue_[i] == mapCellIndex)
        {
            // Item already queued.
            return i;
        }
    }

    // Queue full so nothing to add.
    return QUEUE_FULL;
}

void startHedgeTimer(u8 x, u8 y, u8 roomX, u8 roomY)
{
    u16 cell = mapCellPosition(x, y, roomX, roomY);

    // Check that the cell is a new cell rather than the existing cell.
    // Unless it is an existing cell and the timer has been deactivated.
    if (hedgeCell_ != cell || !hedgeTimerActive)
    {
        // Check if the hedge cell currently has a value.
        if (hedgeCell_ != HEDGE_CELL_UNSET)
        {
            // If it has a value it needs to be redrawn to remove the blended machete graphic from the hedge cell.
            u8 x = (hedgeCell_ / MAP_TILES_Y) % SCREEN_TILES_X;
            u8 y = (hedgeCell_ % MAP_TILES_X) % SCREEN_TILES_Y;
            drawCellAtPosition(x, y);
        }

        // Draws the machete as a blended sprite to indicate that the player is trying to use the machete on a hedge cell.
        cpct_setBlendMode(CPCT_BLEND_AND);
        drawBlendedCell(x, y, CELL_ITEM_FLAG + ITEM_MACHETE);

        // Save the hedge cell that is being destroyed by the machete.
        hedgeCell_ = cell;

        // Reset and start the hedge timer to 0 for the delay waiting for the hedge to be destroyed.
        hedgeTimerCount = 0;
        hedgeTimerActive = TRUE;
    }
}

void updateHedgeCell()
{
    // Transition hedge cell through various hedge states.
    if (mapCells[hedgeCell_] >= CELL_ENVIRONMENT_FLAG + ENVIRONMENT_HEDGE_0 && mapCells[hedgeCell_] <= CELL_ENVIRONMENT_FLAG + ENVIRONMENT_HEDGE_2)
    {
        // Change from hedge cell to default environment.
        u8 x = (hedgeCell_ / MAP_TILES_Y) % SCREEN_TILES_X;
        u8 y = (hedgeCell_ % MAP_TILES_X) % SCREEN_TILES_Y;

        if (mapCells[hedgeCell_] == CELL_ENVIRONMENT_FLAG + ENVIRONMENT_HEDGE_2)
        {
            // Transition hedge state to clear.
            mapCells[hedgeCell_] = CLEAR_CELL;
        }
        else
        {
            // Transition hedge state to next hedge state (0 to 1, 1 to 2).
            ++mapCells[hedgeCell_];
        }

        // Draws the new cell at the specified position.
        drawCellAtPosition(x, y);

        // Play machete sound on transitioning (hitting) the hedge cell.
        playSoundEffect(SFX_MACHETE, CHANNEL_A, 4);

        // Reset hedge cell after update ready for the next hedge the player attacks.
        hedgeCell_ = HEDGE_CELL_UNSET;
    }

    // Remove the machete from the inventory after use.
    inventoryItems &= ~INVENTORY_ITEM_MACHETE;

    // Redraw the hud after inventory update.
    hudRedraw = TRUE;
}

void stopHedgeTimer()
{
    if (hedgeCell_ != HEDGE_CELL_UNSET)
    {
        u8 x = (hedgeCell_ / MAP_TILES_Y) % SCREEN_TILES_X;
        u8 y = (hedgeCell_ % MAP_TILES_X) % SCREEN_TILES_Y;
        drawCellAtPosition(x, y);

        // Reset hedge cell to unset to prevent stopHedgeTimer running when it shouldn't.
        hedgeCell_ = HEDGE_CELL_UNSET;
    }

    hedgeTimerActive = FALSE;
}