#include "Game.h"

// Backup game score to use when a checkpoint is rejected.
u16 nonRestoreGameScore_ = 0;

void runGameSession()
{
    // Initialise levelStart variables to the start of the selected level.
    u8 levelStartX = level_start_position[currentMap][0];
    u8 levelStartY = level_start_position[currentMap][1];

    // Clear everything off the screen.
    cpct_clearScreen(0);

    // Reset game state values for a new game.
    resetGameSessionState();

    // Set the game to running for the game loop.
    gameRunning = TRUE;

    // Initialises arkos 2 sounds.
    initSound();

    // Loop while level state is ready.
    while (!levelState)
    {
        // Starts playing the game music (alternating between the two pieces of music).
        playMusic(currentMap % 2);

        // Resets the level score to the current game score (0 at the start of the game session)
        levelScore = gameScore;

        // Only reset the map and inventory when not loading from a checkpoint.
        if (checkpointX == 255)
        {
            // Initialise the map from the level array.
            initMap();

            // Clear inventory on new level start.
            inventoryItems = 0;

            // Enables the level timer to start the time ticking down.
            enableLevelTimer = TRUE;
        }
        else
        {
            // Display the show restore dialog to allow the player to confirm 
            // that they was to load from the previously saved checkpoint.
            // This allows the player to reject an unfriendly checkpoint and restart the level.
            // The player might also need to reject a checkpoint if the map has changed from crumbling blocks.
            showRestoreDialog = TRUE;

            // If the player dies from running out of time we want them to reset back to the starting amount of time in the level.
            if (levelTimerSecondsRemaining < level_map_time[currentMap])
            {
                levelTimerSecondsRemaining = level_map_time[currentMap];

                // Pause the level timer until the player has decided to restore or not.
                enableLevelTimer = FALSE;
            }

            // Map must be redrawn when loading from checkpoint.
            mapRedraw = TRUE;
        }

        // Initialise the player values, with map position.
        initPlayer(levelStartX, levelStartY, checkpointX != 255);

        // Initialise the projectile state for all projectiles (player and enemy).
        initProjectiles();

        // Draw the hud on the bottom of the screen.
        hudRedraw = TRUE;

        // Initialise the crumble queue to empty before the level starts.
        initCrumbleQueue();

        // Reset the crumbling queue before the level starts
        crumblingTimerActive = FALSE;

        // Reset border before the level starts.
        // This also handles what happens if the map has no artefact pieces.
        cpct_setBorder(HW_BLUE);

        // Run the actual game including the game loop.
        runGame();

        if (levelState == LEVEL_COMPLETE)
        {
            // Add a 100 (shown as 1000) bonus for completing a level without losing health.
            if (player.health >= playerStartingHealth)
            {
                levelScore += 100;
            }

            // Time bonus only applies on normal and up as easy and below get too much time.
            if (chosenDifficulty == BRUTAL_DIFFICULTY)
            {
                levelScore += levelTimerSecondsRemaining * 2;
            }
            else if (chosenDifficulty > EASY_DIFFICULTY)
            {
                // Add the remaining time in the level as a bonus (effectively multiplied by 10).
                levelScore += levelTimerSecondsRemaining;
            }

            // Update the game score to the level score after finishing the level.
            gameScore = levelScore;

            // Backup the game score in case we need it when a checkpoint is rejected.
            nonRestoreGameScore_ = levelScore;

            // Update current map to prepare to move to the next level.
            ++currentMap;

            // Reset the checkpoint variables.
            checkpointX = 255;
            checkpointY = 255;

            // Check if all the levels have been completed.
            if (currentMap == LEVEL_COUNT)
            {
                // If currentMap is the last level then game complete
                levelState = LEVEL_GAME_COMPLETE;
            }
            else
            {
                // Move to the next level.
                levelState = LEVEL_READY;
                gameRunning = TRUE;

                // Update level start position for the next level.
                levelStartX = level_start_position[currentMap][0];
                levelStartY = level_start_position[currentMap][1];
            }
        }
        else if (levelState == LEVEL_DEATH)
        {
            if (playerLives)
            {
                // Player has died reload the current map.
                levelState = LEVEL_READY;
                gameRunning = TRUE;

                // If a checkpoint has been set update the level start variables to that location.
                if (checkpointX != 255)
                {
                    // Checkpoint accepted so redraw the map.
                    mapRedraw = TRUE;

                    // Update the game score to the level score after dying in the level.
                    gameScore = levelScore;

                    // Update level start to checkpoint coordinates.
                    levelStartX = checkpointX;
                    levelStartY = checkpointY;
                }
                else
                {
                    // Make sure that we restart at the start of the level if no checkpoint is loaded.
                    levelStartX = level_start_position[currentMap][0];
                    levelStartY = level_start_position[currentMap][1];
                }
            }
            else
            {
                // Update the game score to the level score after dying part way through the level.
                gameScore = levelScore;

                // Player no longer has any lives so game over.
                levelState = LEVEL_GAME_OVER;
            }
        }
        else if (levelState == LEVEL_EXIT)
        {
            // Update the game score when exiting the level using escape.
            gameScore = levelScore;
        }
    }

    // Reset the border if the player ends the game (win, lose or exit).
    cpct_setBorder(HW_BLUE);

    // Turn off the level timer once the level has ended.
    enableLevelTimer = FALSE;

    // Stop the sound when the level has ended (game over or exit)
    stopSound();
}

void runGame()
{
    u8 logicFlag = 0;

    while (gameRunning)
    {
        // Read input from the player.
        readInput();

        // Only run logic every second time.
        if (logicFlag++)
        {
            // Run the game logic which will change what is displayed.
            gameLogic();

            // Reset flag
            logicFlag = 0;
        }

        // Wait for vsync.
        cpct_waitVSYNC();

        // Draws the game on the screen.
        drawGame();

        // Display the restore dialog to let the user decide if they want to accept
        // the previously activated checkpoint.
        // This enables showing the map under the dialog so the player knowes where they are spawning.
        if (showRestoreDialog)
        {
            // Make sure the dialog is not shown again after the decision is made.
            showRestoreDialog = FALSE;

            if (restoreDialog())
            {
                // Checkpoint accepted so redraw the map.
                mapRedraw = TRUE;

                // Make sure to redraw the player so it doesn't disappear after the map redraw.
                player.redraw = TRUE;

                // Resume the level timer.
                enableLevelTimer = TRUE;
            }
            else
            {
                // Checkpoint rejected so end the session and restart the level (no lives will be lost).
                levelState = LEVEL_DEATH;

                // Clear the checkpoint that was rejected.
                checkpointX = 255;
                checkpointY = 255;

                // Restore game score to the starting level score on checkpoint rejection.
                gameScore = nonRestoreGameScore_;

                // Escapes back to the level loop.
                return;
            }
        }
    }
}

void readInput()
{
    // Reads the game controls from the player.
    readGameControls();

    // Check that the exit key has been pressed.
    if (gameControls.controlFlags & CONTROL_EXIT_FLAG)
    {
        // Pause the level timer.
        enableLevelTimer = FALSE;

        // If the player chooses to press yes exit the game.
        if (exitDialog())
        {
            gameRunning = FALSE;
            levelState = LEVEL_EXIT;
        }
        else
        {
            // Resume the level timer.
            enableLevelTimer = TRUE;

            // Force a map redraw to cleanup after the dialog is closed.
            mapRedraw = TRUE;

            // Redraw the player after drawing the map.
            player.redraw = TRUE;
        }
    }
}

void drawGame()
{
    // Draws the map grid on the screen.
    drawMap();

    // Draw the player on the screen at their current position.
    drawPlayer();

    // Draw the enemies on the screen.
    // If the player is within the enemy backbuffer the player will be drawn without a backbuffer.
    drawEnemies();

    // Draw any projectiles on the screen (player or enemy).
    drawProjectiles();

    // Draw the hud at the bottom of the screen.
    drawHud();
}

void gameLogic()
{
    // Move the player (update velocity).
    movePlayer();

    // Run the logic for the enemies (update position and velocity).
    logicEnemies();

    // Run the logic for the player (update position).
    logicPlayer();

    // Test whether player has collided with any enemies.
    playerEnemyCollision();

    // Run logic for the projectiles after running the player logic.
    logicProjectiles();
}

void generatePasscode(u8 mapNumber, u8* code)
{
    static const u8 offsets[PASS_CODE_LENGTH] = { 4, 17, 7, 23, 8 };
    u8 i = 0;
    u8 letter = mapNumber * 4;

    for (i = 0; i < PASS_CODE_LENGTH; ++i)
    {
        while (letter >= 26)
        {
            letter -= 26;
        }

        *code = letter + 'a';
        letter += offsets[i];

        // Move onto the next element in the array.
        ++code;
    }

    // Add the null terminator to the end of the string.
    *code = 0;
}