#include "Map.h"

// The map grid.
u8 mapCells[MAP_CELL_COUNT];

// The number of enemy entities on the map.
u8 mapEnemyCount;

// The enemy entities on the map.
Enemy mapEnemies[MAX_ENEMY_COUNT];

// Redraw variables.
u8 mapRedraw = TRUE;

// The map index of the current map in the maps array.
u8 currentMap = 0;

// The back buffer where tiles and actors are combined before drawing.
u8 backbuffer[BACKBUFFER_SIZE];

// The width of the back buffer where tiles and actors are combined before drawing.
u8 backbufferWidth = BACKBUFFER_WIDTH;

// The cell type used by default in the map grid.
u8 defaultEnvironmentCellType = 0;

// The cell type (including flag) used when clearing a cell in the map grid.
u8 defaultClearCellType = 0;

// The number of artefact pieces that the player needs to collect to complete the level.
u8 remainingArtefactPieces = 0;

// The cell that the crumbling blocks will crumble to at the end of the crumble cycle.
u8 endCrumbleCell = 0;

void initMap()
{
    // The end of the source level array (arraypointer + size - 1) as a u16 pointer.
    void* sourceEnd = level_map[currentMap] + level_map_size[currentMap] - 1;

    // The end of the destination array (arraypointer + size - 1) as a u16 pointer.
    void* destinationEnd = mapCells + MAP_CELL_COUNT - 1;

    cpct_zx7b_decrunch_s(destinationEnd, sourceEnd);

    // Only load the enemies into the array if there are enemies in the map.
    // Divide the value by the components (x, y, type) to get actual count of enemies.
    mapEnemyCount = (level_enemies_array_size[currentMap] / ENEMY_ARRAY_COMPONENT_COUNT);

    // Only load them into the array if there are enemies in the map.
    if (mapEnemyCount > 0)
    {
        // Ensure the enemies are cleared before they are initialised.
        cpct_memset(mapEnemies, 0, MAX_ENEMY_COUNT * sizeof(Enemy));

        // Initialise the enemies in the map.
        initEnemies(level_enemies[currentMap], level_enemies_array_size[currentMap]);
    }

    // Set the default cell type that will be used for cells in the grid (used when setting cells to empty).
    defaultEnvironmentCellType = level_default_environment_cell[currentMap];

    // Sets the cell type that will be used for cells in the grid when a cell is cleared such as an item being picked up.
    defaultClearCellType = level_clear_cell[currentMap];

    // The number of seconds remaining in the level.
    levelTimerSecondsRemaining = level_map_time[currentMap];

    // Update the palette for either colour or green based on map palette selection.
    updatePalette(level_palette[currentMap]);

    // The number of artefacts the player needs to collect in the level.
    remainingArtefactPieces = level_artefact_piece_count[currentMap];

    // The cell that the crumbling blocks will crumble to for the level.
    endCrumbleCell = level_crumble_cell[currentMap];

    // Enable map redraw.
    mapRedraw = TRUE;
}

void drawMap()
{
    if (mapRedraw)
    {
        // Draw the content of the map grid.
        u8 x;
        u8 y;
        u8 cell;

        // Position is determined by how much the screen has scrolled to show the current room.
        u16 position = mapCellPosition(0, 0, currentRoomX, currentRoomY);
        u8* cellPointer = mapCells + position;

        // Position is the element in the map tile array.
        for (x = 0; x < SCREEN_TILES_X; ++x)
        {
            for (y = 0; y < SCREEN_TILES_Y; ++y)
            {
                cell = *cellPointer;

                drawCell(x, y, cell);

                ++cellPointer;
            }

            // Move to the start of the next column (half the column height in a 2x2 screen map), 2/3 of the column height in a 3x3 map.
            cellPointer += 24;
        }

        // Force redraw of the enemies in the map.
        markEnemiesForRedraw();

        mapRedraw = FALSE;
    }
}

void drawCell(u8 x, u8 y, u8 cell)
{
    u8* mapMemory = cpct_getScreenPtr(CPCT_VMEM_START, xPositions[x], yPositions[y]);
    u8* graphic = graphicFromCell(cell);

    cpct_drawSprite(graphic, mapMemory, TILE_WIDTH_BYTES, TILE_HEIGHT);
}

void drawBlendedCell(u8 x, u8 y, u8 cell)
{
    u8* mapMemory = cpct_getScreenPtr(CPCT_VMEM_START, xPositions[x], yPositions[y]);
    u8* graphic = graphicFromCell(cell);

    cpct_drawSpriteBlended(mapMemory, TILE_HEIGHT, TILE_WIDTH_BYTES, graphic);
}

u16 mapCellPosition(u8 x, u8 y, u8 roomX, u8 roomY)
{
    return (x + (SCREEN_TILES_X * roomX)) * MAP_TILES_Y + (y + (SCREEN_TILES_Y * roomY));
}

u8 mapCell(u8 x, u8 y, u8 roomX, u8 roomY)
{
    return mapCells[mapCellPosition(x, y, roomX, roomY)];
}

void addSpriteToBackbuffer(u8 x, u8 y, u8* sprite)
{
    // Determine the position within the backbuffer that the new sprite should be added to.
    u8 position = y * backbufferWidth + x;
    u8* mapMemory = backbuffer + position;

    cpct_drawToSpriteBuffer(backbufferWidth, mapMemory, TILE_WIDTH_BYTES, TILE_HEIGHT, sprite);
}

void drawBackbuffer(u8 x, u8 y)
{
    u8* memory = cpct_getScreenPtr(CPCT_VMEM_START, xPositions[x], yPositions[y]);

    // When at the bottom of the screen we only draw half the backbuffer height.
    u8 height = y < 11 ? BACKBUFFER_HEIGHT : TILE_HEIGHT;

    cpct_drawSprite(backbuffer, memory, backbufferWidth, height);
}

void populateBackbufferFromMap(u8 cellX, u8 cellY)
{
    // Get the array position from mapCells for (x,y).
    u16 cellPosition = mapCellPosition(cellX, cellY, currentRoomX, currentRoomY);
    u8* cell = mapCells + cellPosition;

    // When cellX is 19 the backbuffer is 1x2.
    addSpriteToBackbuffer(0, 0, graphicFromCell(*cell));

    // Move cell to (x,y+1).
    ++cell;

    addSpriteToBackbuffer(0, TILE_HEIGHT, graphicFromCell(*cell));

    if (cellX != 19)
    {
        // Normally backbuffer is 2x2.
        // Move cell to (x+1,y+1).
        cell += MAP_TILES_Y;
        addSpriteToBackbuffer(TILE_WIDTH_BYTES, TILE_HEIGHT, graphicFromCell(*cell));

        // Move cell to (x+1,y).
        --cell;
        addSpriteToBackbuffer(TILE_WIDTH_BYTES, 0, graphicFromCell(*cell));
    }
}

u8* graphicFromCell(u8 cell)
{
    // The cell type without the flag.
    u8 cellType = cell & CELL_GFX_RANGE;

    if (cell & CELL_ITEM_FLAG)
    {
        // Draw item (-1 because the graphicset will start at 0 index but the type will start at 1).
        return itemGraphics[cellType - 1];
    }
    else if (cell & CELL_INTERACTABLE_FLAG)
    {
        // Draw interactable (-1 because the graphicset will start at 0 index but the type will start at 1).
        return interactableGraphics[cellType - 1];
    }
    else if (cell & CELL_ENVIRONMENT_FLAG)
    {
        if (cellType == ENVIRONMENT_HIDDEN_TRIGGER)
        {
            // Check if the player is wearing classes.
            if (inventoryContainsItem(ITEM_GLASSES))
            {
                // Draw the environment hidden trigger cell as the player is wearing glasses.
                return environmentGraphics[ENVIRONMENT_HIDDEN_TRIGGER - 1];
            }
            else
            {
                // Draw the default environment cell as the hidden trigger is hidden from the player.
                return environmentGraphics[DEFAULT_CELL - CELL_ENVIRONMENT_FLAG - 1];
            }
        }
        else if (cellType >= ENVIRONMENT_WALL_FLOW && cellType < ENVIRONMENT_GATE_OPEN)
        {
            // Draw the default environment cell as the cell is hidden from the player.
            return environmentGraphics[DEFAULT_CELL - CELL_ENVIRONMENT_FLAG - 1];
        }
        else if (cellType == ENVIRONMENT_GATE_OPEN)
        {
            // Open gate needs to be drawn as clear cell as I am using gates in the lava level
            // and when you open them it should be save instead of lava (although the lava doesn't damage you).
            return environmentGraphics[CLEAR_CELL - CELL_ENVIRONMENT_FLAG - 1];
        }
        else if (cellType == ENVIRONMENT_HIDDEN_SOLID)
        {
            // Draw the default environment cell for a hidden solid.
            return environmentGraphics[DEFAULT_CELL - CELL_ENVIRONMENT_FLAG - 1];
        }

        // Draw environment cell (-1 because the graphicset will start at 0 index but the type will start at 1).
        return environmentGraphics[cellType - 1];
    }
    else if (cell != CELL_EMPTY)
    {
        // Draw tile (-1 because the graphicset will start at 0 index but the type will start at 1).
        return tileGraphics[cellType - 1];
    }
    else
    {
        // Draw default environment cell for empty cells.
        return environmentGraphics[DEFAULT_CELL - CELL_ENVIRONMENT_FLAG - 1];
    }
}

void updateMapCell(u8 x, u8 y, u8 roomX, u8 roomY, u8 cell)
{
    mapCells[mapCellPosition(x, y, roomX, roomY)] = cell;
}

void drawCellAtPosition(u8 x, u8 y)
{
    u8 cell = mapCell(x, y, currentRoomX, currentRoomY);

    drawCell(x, y, cell);
}

void updateMapCellRedraw(u8 x, u8 y, u8 roomX, u8 roomY, u8 cell)
{
    mapCells[mapCellPosition(x, y, roomX, roomY)] = cell;

    // We will assume it is in the current room and if not it will draw something else.
    drawCellAtPosition(x, y);
}

u8 cellContainsType(u8 x, u8 y, u8 type)
{
    // Retrieve the cell in the current room.
    u8 cell = mapCell(x, y, currentRoomX, currentRoomY);

    // Check if the cell is a particular type of cell.
    if (cell == type)
    {
        return TRUE;
    }

    return FALSE;
}