#include "Menu.h"

// Colour mode menu item text.
const u8* colourModeText_;
const u8 GREEN_MODE_TEXT_[14] = "6. green mode";
const u8 COLOUR_MODE_TEXT_[15] = "6. colour mode";

// Sound menu item text.
const u8* soundModeText_;
const u8 SOUND_OFF_TEXT_[13] = "7. sound off";
const u8 SOUND_ON_TEXT_[12] = "7. sound on";

// Difficulty level menu item text.
const u8 PRACTICE_DIFFICULTY_TEXT_[9] = "practice";
const u8 EASY_DIFFICULTY_TEXT_[5] = "easy";
const u8 NORMAL_DIFFICULTY_TEXT_[7] = "normal";
const u8 HARD_DIFFICULTY_TEXT_[5] = "hard";
const u8 BRUTAL_DIFFICULTY_TEXT_[7] = "brutal";
u8* const difficultyTextOptions_[5] = { PRACTICE_DIFFICULTY_TEXT_, EASY_DIFFICULTY_TEXT_, NORMAL_DIFFICULTY_TEXT_, HARD_DIFFICULTY_TEXT_, BRUTAL_DIFFICULTY_TEXT_ };

// Menu item number controls.
#define START_GAME_KEY_ Key_1
#define PASS_CODE_KEY Key_2
#define REBIND_CONTROLS_KEY_ Key_3
#define DIFFICULTY_KEY_ Key_4
#define HIGH_SCORES_KEY_ Key_5
#define COLOUR_MODE_KEY_ Key_6
#define SOUND_MODE_KEY_ Key_7

// The current mode for playing sound (on, sound effects, music and sound effects).
u8 soundMode = SOUND_MODE_OFF;

// High score variables.
// The number of high score entries.
#define SCORE_COUNT_ 5

// The maximum length that a high score name can be.
#define MAX_NAME_LENGTH_ 10

// The high score values in the high score list.
u16 highScores_[SCORE_COUNT_];

// The names for the high score values in the high score list.
u8 highScoreNames_[SCORE_COUNT_][11];

// User entered the wrong pass code.
#define INVALID_PASS_CODE_ 255

void runMenu()
{
    // Default graphics mode is colour mode.
    colourModeText_ = COLOUR_MODE_TEXT_;

    // Default to sound enabled but not turned on.
    soundModeText_ = SOUND_ON_TEXT_;
    soundMode = SOUND_MODE_OFF;

    // Default difficulty is normal difficulty.
    chosenDifficulty = NORMAL_DIFFICULTY;

    // Set the default high score entries.
    initHighScores();

    // Set the default keys for the player.
    setDefaultKeys();

    // Draw the menu on screen.
    drawMenu();

    // Main menu loop.
    while (1)
    {
        if (cpct_isKeyPressed(START_GAME_KEY_))
        {
            // Game should default to starting at the first level.
            currentMap = 0;

            // Player playing without code so is eligible for game complete bonus.
            passcodeUsed = FALSE;

            // Runs the full game session and displays victory or defeat based on outcome (starting at level 0).
            runFullGame();

            // Redraw the menu after finishing the game session.
            drawMenu();
        }
        else if (cpct_isKeyPressed(PASS_CODE_KEY))
        {
            // Attempt to select a map using the pass code selection screen.
            u8 selectedMap = enterPasscode();

            if (selectedMap != INVALID_PASS_CODE_)
            {
                currentMap = selectedMap;

                // Player used a pass code and will not be eligible for game complete bonus.
                passcodeUsed = TRUE;

                // Runs the full game session and displays victory or defeat based on outcome (starting at the selected level).
                runFullGame();
            }

            // Redraw the menu after finishing the game session or getting the wrong code.
            drawMenu();
        }
        else if (cpct_isKeyPressed(REBIND_CONTROLS_KEY_))
        {
            // Display rebind controls menu which asks the user for their chosen controls.
            rebindControls();

            // Redraw the menu after leaving rebind controls menu.
            drawMenu();
        }
        else if (cpct_isKeyPressed(DIFFICULTY_KEY_))
        {
            chooseDifficulty();

            // Redraw the menu after updating the menu item text.
            drawMenu();
        }
        else if (cpct_isKeyPressed(HIGH_SCORES_KEY_))
        {
            displayHighScores();

            // Redraw the menu after leaving the high scores.
            drawMenu();
        }
        else if (cpct_isKeyPressed(COLOUR_MODE_KEY_))
        {
            switchColourMode();

            // Redraw the menu after updating the menu item text.
            drawMenu();
        }
        else if (cpct_isKeyPressed(SOUND_MODE_KEY_))
        {
            switchSoundMode();

            // Redraw the menu after updating the menu item text.
            drawMenu();
        }
    }
}

void drawMenu()
{
    // Clear the screen before drawing the menu.
    cpct_clearScreen(0);

    // Game title.
    drawString(3, 5, "fitzroy and the confounding labyrinth");

    // Menu item choices.
    drawString(20, 45, "1. start game");
    drawString(20, 60, "2. pass code");
    drawString(20, 76, "3. rebind controls");
    drawString(20, 90, "4. difficulty");
    drawString(48, 90, difficultyTextOptions_[chosenDifficulty]);
    drawString(20, 105, "5. high scores");
    drawString(20, 120, colourModeText_);
    drawString(20, 135, soundModeText_);

    // Draw the game version number on the bottom right corner of the screen.
    drawString(70, 180, GAME_VERSION);

    // Draw the build number on the bottom right corner of the screen.
    drawString(70, 190, BUILD_NUMBER);
}

void chooseDifficulty()
{
    // Switch between difficulties that will update the menu text.
    if (chosenDifficulty == BRUTAL_DIFFICULTY)
    {
        chosenDifficulty = PRACTICE_DIFFICULTY;
    }
    else
    {
        ++chosenDifficulty;
    }

#ifdef WINCPCTELERA
    // Add a delay to make it easier to switch modes with wincpctelera.
    wincpct_wait(200);
#endif
}

void switchColourMode()
{
    // Update menu item text to reflect the chosen colour mode.
    if (greenMode)
    {
        greenMode = FALSE;
        colourModeText_ = COLOUR_MODE_TEXT_;
    }
    else
    {
        greenMode = TRUE;
        colourModeText_ = GREEN_MODE_TEXT_;
    }

#ifdef WINCPCTELERA
    // Add a delay to make it easier to switch modes with wincpctelera.
    wincpct_wait(200);
#endif
}

void rebindControls()
{
    // Clear the screen so that we can show control text.
    cpct_clearScreen(0);

    // Enter controls title.
    drawString(15,8, "enter controls");

    // Draw the text indicating which control is being set.
    // Make sure no key is currently down.
    // Set the key for the next control (this is done for each of control below).
    drawString(25, 30, "up");
    gameControls.up = getKey();

    drawString(25, 45, "down");
    gameControls.down = getKey();

    drawString(25, 60, "left");
    gameControls.left = getKey();

    drawString(25, 75, "right");
    gameControls.right = getKey();

    drawString(25, 90, "action");
    gameControls.use = getKey();

    drawString(25, 105, "exit");
    gameControls.exit = getKey();
}

cpct_keyID getKey()
{
    // From Ronaldo and Arnaud on the cpcwiki forum.
    u8* keyStatus = cpct_keyboardStatusBuffer;
    u8 i;

    // Wait for any previous key presses to clear first.
    while (cpct_isAnyKeyPressed());

    // Then wait for the desired key press.
    while (!cpct_isAnyKeyPressed());

    for (i = 0; i < 10; ++i)
    {
        cpct_keyID keypressed = *keyStatus++ ^ 0xFF;

        if (keypressed)
        {
            return (keypressed << 8) + i;
        }
    }

    return 0;
}

void switchSoundMode()
{
    if (soundMode == SOUND_MODE_SOUND_DISABLED)
    {
        soundMode = SOUND_MODE_OFF;
        soundModeText_ = SOUND_ON_TEXT_;
    }
    else
    {
        soundMode = SOUND_MODE_SOUND_DISABLED;
        soundModeText_ = SOUND_OFF_TEXT_;
    }

#ifdef WINCPCTELERA
    // Add a delay to make it easier to switch modes with wincpctelera.
    wincpct_wait(200);
#endif
}

void runFullGame()
{
    // Runs the full game session.
    runGameSession();

    // Displays either game over or victory screen depending on if the player succeeded or not.
    displayEndScreen();

    // Update high scores.
    updateHighScores();

    // Display high scores after entering high score.
    displayHighScores();
}

void displayEndScreen()
{
    // Wait for any previous key presses to clear first.
    while (cpct_isAnyKeyPressed());

    // Clear the screen
    cpct_clearScreen(0);

    if (levelState == LEVEL_GAME_COMPLETE)
    {
        // Draw victorious text on screen.
        u8* memory = cpctm_screenPtr(CPCT_VMEM_START, 24, 16);
        drawString(30, 44, "victorious");

        // Draw staff on the screen.
        cpct_drawSprite(G_item_staff_top, memory, TILE_WIDTH_BYTES, TILE_HEIGHT);
        memory += 160;
        cpct_drawSprite(G_item_staff_middle, memory, TILE_WIDTH_BYTES, TILE_HEIGHT);
        memory += 160;
        cpct_drawSprite(G_item_staff_middle, memory, TILE_WIDTH_BYTES, TILE_HEIGHT);
        memory += 160;
        cpct_drawSprite(G_item_staff_bottom, memory, TILE_WIDTH_BYTES, TILE_HEIGHT);

        // Display bonus text heading.
        drawString(34, 96, "bonus");

        // Only award bonuses if the player played without a pass code.
        if (!passcodeUsed)
        {
            // 1000 (after the extra 0 is displayed) points multiplied by difficulty for winning the game.
            gameScore += (chosenDifficulty + 1) * 100;

            // Starting memory position for bonuses.
            memory = cpctm_screenPtr(CPCT_VMEM_START, 34, 112);

            // TODO handle if we support adding additional lives with points.
            // 2000 (after the extra 0 is displayed) points multiplied by difficulty for not losing a life.
            if (playerLives == startingLivesCount)
            {
                gameScore += (chosenDifficulty + 1) * 200;

                // Draw player graphic if player gets the life bonus.
                cpct_drawSprite(G_player_right, memory, TILE_WIDTH_BYTES, TILE_HEIGHT);
            }

            // 1000 points if the player collects all the coins (assumed one per level).
            if (coinCount == LEVEL_COUNT)
            {
                gameScore += 100;

                // Display a coin graphic if player the player gets the coin bonus.
                memory += 6;
                cpct_drawSprite(G_item_coin, memory, TILE_WIDTH_BYTES, TILE_HEIGHT);
            }
        }

        // Play fanfare music (fin2) when the player gets to the victory screen.
        initSound();
        playMusic(MUSIC_FIN2);
    }
    else
    {
        static u8 generatedCode[6];

        // Display Game Over.
        drawString(30, 20, "game over");

        // Display the pass code for the level the player got up to before dying.
        drawString(30, 40, "pass code");
        generatePasscode(currentMap, generatedCode);
        drawString(34, 50, generatedCode);

        // Play died music (died2) when the player gets to the victory screen.
        initSound();
        playMusic(MUSIC_DIED2);
    }

    // Wait for a key press before going back to the main menu.
    while (!cpct_isAnyKeyPressed());

    // Stop the music after the player leaves the screen.
    stopSound();
}

void displayHighScores()
{
    u8 i = 0;
    u8 scoreStart = 25;

    cpct_clearScreen(0);

    // High scores title.
    drawString(25, 5, "high scores");

    // Each of the entires in the high score list.
    for (i = 0; i < SCORE_COUNT_; ++i)
    {
        u8 y = 45 + i * 15;

        // Draw the name of the player from the names array.
        drawString(40, y, highScoreNames_[i]);

        // Draw the score with the zeros (there is an extra 0 so that scores can be bigger than the are actually stored).
        drawScore(highScores_[i], scoreStart, y);
    }

    getKey();
}

void initHighScores()
{
    u8 i = 0;

    for (i = 0; i < SCORE_COUNT_; ++i)
    {
        highScores_[i] = 10;
        cpct_memcpy(highScoreNames_[i], "cpc\0", 4);
    }
}

void updateHighScores()
{
    // Check that the player got a higher score than the bottom score.
    if (gameScore > highScores_[4])
    {
        // Let the player enter their name.
        static u8 scoreName[11];
        u8 i = 0;

        // Set scoreName to the null terminator to end the string.
        cpct_memset(scoreName, '\0', 11);

        cpct_clearScreen(0);
        drawString(15, 5, "new high score");

        // Draw the score with padding of zeroes.
        drawScore(gameScore, 45, 5);
        drawString(20, 25, "enter your name");

        // Enter the name of the user for the high scores.
        enterWord(scoreName, MAX_NAME_LENGTH_);

        // Update score values.
        highScores_[4] = gameScore;
        cpct_memcpy(highScoreNames_[4], scoreName, 11);

        for (i = 4; i > 0; --i)
        {
            if (highScores_[i] > highScores_[i - 1])
            {
                highScores_[i] = highScores_[i - 1];
                cpct_memcpy(highScoreNames_[i], highScoreNames_[i - 1], 11);

                highScores_[i - 1] = gameScore;
                cpct_memcpy(highScoreNames_[i - 1], scoreName, 11);
            }
        }
    }
}

void enterWord(u8* word, u8 maxLength)
{
    // Default the chosen key to whatever.
    cpct_keyID chosenKey = Key_0;
    u8 letterCount = 0;
    u8 letter = 0;

    do
    {
        // Get the next character from the user.
        chosenKey = getKey();

        // Del to move back to the previous key.
        if (chosenKey == Key_Del)
        {
            // Go back to the previous letter.
            if (letterCount)
            {
                --letterCount;
                --word;

                // Draw letter showing a space where it used to be a letter.
                drawString(25 + letterCount * 2, 50, " ");
            }
        }

        // Make sure the key is in the correct range and the word is not too long.
        else if (letterCount < maxLength)
        {
            // Convert the chosen key (cpct_keyID to a character to an ASCII character).
            switch (chosenKey)
            {
            case Key_A:
                letter = 'a';
                break;

            case Key_B:
                letter = 'b';
                break;

            case Key_C:
                letter = 'c';
                break;

            case Key_D:
                letter = 'd';
                break;

            case Key_E:
                letter = 'e';
                break;

            case Key_F:
                letter = 'f';
                break;

            case Key_G:
                letter = 'g';
                break;

            case Key_H:
                letter = 'h';
                break;

            case Key_I:
                letter = 'i';
                break;

            case Key_J:
                letter = 'j';
                break;

            case Key_K:
                letter = 'k';
                break;

            case Key_L:
                letter = 'l';
                break;

            case Key_M:
                letter = 'm';
                break;

            case Key_N:
                letter = 'n';
                break;

            case Key_O:
                letter = 'o';
                break;

            case Key_P:
                letter = 'p';
                break;

            case Key_Q:
                letter = 'q';
                break;

            case Key_R:
                letter = 'r';
                break;

            case Key_S:
                letter = 's';
                break;

            case Key_T:
                letter = 't';
                break;

            case Key_U:
                letter = 'u';
                break;

            case Key_V:
                letter = 'v';
                break;

            case Key_W:
                letter = 'w';
                break;

            case Key_X:
                letter = 'x';
                break;

            case Key_Y:
                letter = 'y';
                break;

            case Key_Z:
                letter = 'z';
                break;

            case Key_Space:
                letter = ' ';
                break;
            }

            if (letter)
            {
                // Array that is used for drawing the current letter on screen.
                static u8 chosenCharacters[2];
                chosenCharacters[1] = '\0';

                // Add the chosen letter to the string that makes up the word that the user is entering.
                *word = letter;

                // Save the letter to chosen characters for drawing.
                chosenCharacters[0] = letter;

                // Draw letter, the one just selected.
                drawString(25 + letterCount * 2, 50, chosenCharacters);

                ++word;
                ++letterCount;

                // Make sure the chosen character is reset so the previous character
                // does not get drawn if the user chooses an invalid character.
                letter = 0;
            }
        }
    } while (chosenKey != Key_Return && chosenKey != Key_Enter);
}

u8 enterPasscode()
{
    static u8 passcode[6];
    static u8 generatedCode[6];
    u8 i = 0;
    u8 j = 0;
    cpct_memset(passcode, '\0', 6);

    // Clear the screen.
    cpct_clearScreen(0);

    // Instruct the user to enter the code
    drawString(15, 25, "enter pass code");

    // Get the user to enter a possible pass code.
    enterWord((u8*)passcode, PASS_CODE_LENGTH);

    // Compare user selection with existing pass codes to see if they are valid.
    for (i = 0; i < LEVEL_COUNT; ++i)
    {
        u8 invalid = FALSE;
        generatePasscode(i, generatedCode);

        for (j = 0; j < PASS_CODE_LENGTH; ++j)
        {
            invalid += (generatedCode[j] != passcode[j]);
        }

        if (!invalid)
        {
            // Code valid, return map number matching selected pass code.
            return i;
        }
    }

    // User selected pass code does not match.
    return INVALID_PASS_CODE_;
}