#include "Player.h"

// Global player object.
Player player;

// Machete activate flag used to ensure the hedge timer can only update the hedge state if the player is actively attacking the hedge.
u8 macheteActive_ = FALSE;

void initPlayer(u8 x, u8 y, u8 loadingCheckpoint)
{
    // Initialise actor variables.
    initActor((Actor*)&player, x, y, ACTOR_TYPE_PLAYER);

    // The current room is the room that the player is in.
    currentRoom = player.object.position.room;
    currentRoomX = currentRoom % 3;
    currentRoomY = currentRoom / 3;

    // Currently the player graphic defaults to right.
    player.direction = DIRECTION_RIGHT;
    player.previousDirection = DIRECTION_RIGHT;
    player.graphicIndex = PLAYER_GRAPHIC_RIGHT;

    // Player item values will only be reset on a normal load.
    // When loading from checkpoint these values should carry over.
    if (loadingCheckpoint == FALSE)
    {
        // Ammo starts at zero.
        player.ammo = 0;
    }

    // Default to health set by difficulty.
    player.health = playerStartingHealth;
}

void drawPlayer()
{
    if (player.redraw)
    {
        // Draws the player using the actor struct.
        drawActor((Actor*)&player, playerGraphics[player.graphicIndex]);

        if (roomChange)
        {
            roomChange = FALSE;
        }

        // Set redraw to false ready for the next change when it will be set to true.
        player.redraw = FALSE;
    }
}

void movePlayer()
{
    // Update the player velocity based on controls currently down.
    if (gameControls.controlFlags & CONTROL_LEFT_FLAG)
    {
        player.object.velocity.x = -1;
        player.direction = DIRECTION_LEFT;
    }

    if (gameControls.controlFlags & CONTROL_RIGHT_FLAG)
    {
        player.object.velocity.x = 1;
        player.direction = DIRECTION_RIGHT;
    }

    if (gameControls.controlFlags & CONTROL_UP_FLAG)
    {
        player.object.velocity.y = -2;
        player.direction = DIRECTION_UP;
    }

    if (gameControls.controlFlags & CONTROL_DOWN_FLAG)
    {
        player.object.velocity.y = 2;
        player.direction = DIRECTION_DOWN;
    }
}

void logicPlayer()
{
    // Save player.inWater to determine if the player has changed from in and out of water.
    // This is used for determining if the player graphic index should be updated.
    u8 previousInWater = player.inWater;

    // The tiles that the player currently occupies.
    u8 occupiedXTileLeft = (player.object.position.x) >> 2;
    u8 occupiedXTileRight = (player.object.position.x + TILE_WIDTH_BYTES - 1);
    u8 occupiedYTileBottom = (player.object.position.y + TILE_HEIGHT - 1);
    u8 occupiedYTileTop = (player.object.position.y) >> 4;
    occupiedXTileRight >>= 2;
    occupiedYTileBottom >>= 4;

    // Reset machete active flag ready to be activated if the player is using the machete on a hedge.
    macheteActive_ = FALSE;

    // The use key is activated by the player when they want to use an interactable.
    if (!useKeyFlag && gameControls.controlFlags & CONTROL_USE_FLAG)
    {
        // Activate any interactible in one of the squares the player occupies.
        u8 interactableActivated = activateInteractable(occupiedXTileLeft, occupiedYTileTop, currentRoomX, currentRoomY, 0);

        // Set use flag to ensure the use key cannot be held down.
        // This prevents an interactable being activated continuously with the use key.
        useKeyFlag = TRUE;

        // Check that the cells are not the same otherwise we get multiple activations.
        if (occupiedYTileTop != occupiedYTileBottom)
        {
            interactableActivated += activateInteractable(occupiedXTileLeft, occupiedYTileBottom, currentRoomX, currentRoomY, 0);
        }

        if (occupiedXTileLeft != occupiedXTileRight)
        {
            interactableActivated += activateInteractable(occupiedXTileRight, occupiedYTileTop, currentRoomX, currentRoomY, 0);

            if (occupiedYTileTop != occupiedYTileBottom)
            {
                interactableActivated += activateInteractable(occupiedXTileRight, occupiedYTileBottom, currentRoomX, currentRoomY, 0);
            }
        }

        // Check the player has not activated an interactable with the user key.
        if (!interactableActivated)
        {
            // Check if the player is carrying a block.
            if (inventoryItems & INVENTORY_ITEM_BLOCK)
            {
                // Only drop the block item if they are carrying the block.
                // You can activate an interactable if carrying a block.
                // The block is dropped left of the direction the player is facing.
                if (player.direction == DIRECTION_UP)
                {
                    dropBlockItem(occupiedXTileLeft, occupiedYTileTop - 1);
                }
                else if (player.direction == DIRECTION_RIGHT)
                {
                    dropBlockItem(occupiedXTileRight + 1, occupiedYTileTop);
                }
                else if (player.direction == DIRECTION_LEFT)
                {
                    dropBlockItem(occupiedXTileLeft - 1, occupiedYTileBottom);
                }
                else
                {
                    dropBlockItem(occupiedXTileRight, occupiedYTileBottom + 1);
                }
            }
            else
            {
                // Use and shoot are shared, only shoot if there is no interactable
                // and not carrying a block.
                spawnProjectile();
            }
        }
    }

    // Save player current position before updating position based on velocity.
    // If you do this after checking velocity it results in fitzroy only being partially drawn (or not at all) on screen.
    player.object.previousPosition.x = player.object.position.x;
    player.object.previousPosition.y = player.object.position.y;

    // Check the player is moving (the block could set velocity to 0).
    if (player.object.velocity.y || player.object.velocity.x)
    {
        // Save the link index before resetting it.
        u8 currentLinkX = player.activatedLinkX;
        u8 currentLinkY = player.activatedLinkY;

        // The tiles the player occupies following a change in velocity.
        u8 futureXTileLeft = (player.object.position.x + player.object.velocity.x) >> 2;
        u8 futureXTileRight = (player.object.position.x + TILE_WIDTH_BYTES - 1 + player.object.velocity.x);
        u8 futureYTileBottom = (player.object.position.y + TILE_HEIGHT - 1 + player.object.velocity.y);
        u8 futureYTileTop = (player.object.position.y + player.object.velocity.y) >> 4;
        futureXTileRight >>= 2;
        futureYTileBottom >>= 4;

        // Reset in water flag which is used by sharks.
        // Only set to false when the player has velocity otherwise it will be set to false
        // when the player stops moving in the water.
        player.inWater = FALSE;

        // Reset link X to determine if the player has moved causing the player to move off the source of the link.
        player.activatedLinkX = NO_LINK_FOUND;

        if (player.object.velocity.x > 0)
        {
            // Player going right.
            if (player.object.position.x + player.object.velocity.x <= SCREEN_X - TILE_WIDTH_BYTES)
            {
                if (processCell(futureXTileRight, occupiedYTileTop) == CELL_EMPTY && 
                    processCell(futureXTileRight, occupiedYTileBottom) == CELL_EMPTY)
                {
                    player.object.position.x += player.object.velocity.x;
                }
            }
            else
            {
                // Check that there is a room on the right (not in room 2).
                if (currentRoomX < 2)
                {
                    // Check that the first cell on the room to the right is empty.
                    if (processCellInRoom(0, occupiedYTileTop, currentRoomX + 1, currentRoomY) == CELL_EMPTY &&
                        processCellInRoom(0, occupiedYTileBottom, currentRoomX + 1, currentRoomY) == CELL_EMPTY)
                    {
                        // Update player position for the new room.
                        moveRoomX(0, 1);
                    }
                }
            }
        }
        else if (player.object.velocity.x < 0)
        {
            // Player going left.
            if (player.object.position.x + player.object.velocity.x >= 0)
            {
                if (processCell(futureXTileLeft, occupiedYTileTop) == CELL_EMPTY && 
                    processCell(futureXTileLeft, occupiedYTileBottom) == CELL_EMPTY)
                {
                    player.object.position.x += player.object.velocity.x;
                }
            }
            else
            {
                // Check that there is a room on the left (not in room 0).
                if (currentRoomX)
                {
                    // Check that the last cell on the room to the left is empty.
                    if (processCellInRoom(SCREEN_TILES_X - 1, occupiedYTileTop, currentRoomX - 1, currentRoomY) == CELL_EMPTY &&
                        processCellInRoom(SCREEN_TILES_X - 1, occupiedYTileBottom, currentRoomX - 1, currentRoomY) == CELL_EMPTY)
                    {
                        // Update player position for the new room.
                        moveRoomX(SCREEN_X - TILE_WIDTH_BYTES, -1);
                    }
                }
            }
        }

        // Update the occupied position with the new x position values.
        // This prevents the player from clipping through tiles.
        occupiedXTileLeft = (player.object.position.x) >> 2;
        occupiedXTileRight = (player.object.position.x + TILE_WIDTH_BYTES - 1);
        occupiedXTileRight >>= 2;

        if (player.object.velocity.y > 0)
        {
            // Player going down.
            // Check whether player next position below is on screen.
            if (player.object.position.y + player.object.velocity.y <= SCREEN_Y_NO_HUD - TILE_HEIGHT)
            {
                if (processCell(occupiedXTileLeft, futureYTileBottom) == CELL_EMPTY && 
                    processCell(occupiedXTileRight, futureYTileBottom) == CELL_EMPTY)
                {
                    player.object.position.y += player.object.velocity.y;
                }
            }
            else
            {
                // Check that the first cell in the room below is empty.
                if (currentRoomY != 2 &&
                    processCellInRoom(occupiedXTileLeft, 0, currentRoomX, currentRoomY + 1) == CELL_EMPTY &&
                    processCellInRoom(occupiedXTileRight, 0, currentRoomX, currentRoomY + 1) == CELL_EMPTY)
                {
                    // Move the player to the room below the current one.
                    moveRoomBelow();
                }
            }
        }
        else if (player.object.velocity.y < 0)
        {            
            // Going up.
            // Check whether player next position above is on screen.
            if (player.object.position.y + player.object.velocity.y >= 0)
            {
                if (processCell(occupiedXTileLeft, futureYTileTop) == CELL_EMPTY && processCell(occupiedXTileRight, futureYTileTop) == CELL_EMPTY)
                {
                    player.object.position.y += player.object.velocity.y;
                }
            }
            else
            {
                // Check that the last cell in the room above is empty.
                if (currentRoomY != 0 &&
                    processCellInRoom(occupiedXTileLeft, SCREEN_TILES_Y - 1, currentRoomX, currentRoomY - 1) == CELL_EMPTY &&
                    processCellInRoom(occupiedXTileRight, SCREEN_TILES_Y - 1, currentRoomX, currentRoomY - 1) == CELL_EMPTY)
                {
                    // Moves the player to the room above the current room.
                    moveRoomAbove();
                }
            }
        }

        // Reset velocity after movement.
        player.object.velocity.y = 0;
        player.object.velocity.x = 0;

        // If the player is not use the machete then the hedge timer is stopped.
        if (!macheteActive_)
        {
            // This will stop the hedge timer and remove the blended sprite used as an indicator that the machete is active.
            stopHedgeTimer();
        }

        // Determine whether the link should be activated or deactivated based on whether the player is moving onto the plate or off it.
        activateActorLink((Actor*)&player, currentLinkX, currentLinkY);
    }

    // Redraw the player at the new position if they have moved.
    if (player.object.position.x != player.object.previousPosition.x || player.object.position.y != player.object.previousPosition.y)
    {
        player.redraw = TRUE;
    }

    // Change the player graphic if the direction has changed or the player is in or out of water.
    if (player.direction != player.previousDirection || previousInWater != player.inWater)
    {
        if (player.direction == DIRECTION_RIGHT)
        {
            player.graphicIndex = PLAYER_GRAPHIC_RIGHT + 4 * player.inWater;
        }
        else if (player.direction == DIRECTION_LEFT)
        {
            player.graphicIndex = PLAYER_GRAPHIC_LEFT + 4 * player.inWater;
        }
        else if (player.direction == DIRECTION_UP)
        {
            player.graphicIndex = PLAYER_GRAPHIC_UP + 4 * player.inWater;
        }
        else
        {
            player.graphicIndex = PLAYER_GRAPHIC_DOWN + 4 * player.inWater;
        }

        // Update previous direction so we can detection future direction changes.
        player.previousDirection = player.direction;
    }

    // If a spike was detected in any of the player movement it is now necessary
    // to check if the player actually collided with the spike.
    // This is necessary because the processCell code does not know whether it is
    // handling the player's future or current position which could end up being blocked
    // by a tile or other solid cell.
    if (player.spikeDetected)
    {
        processSpikeCollision();
    }

    // Player occupies only chasm cells then they die.
    // If there is a room change we want the map to redraw before the player dies.
    // If the player moves fast they might get over the chasm\water but this is an acceptable compromise.
    if (!roomChange && player.environmentCells)
    {
        if (actorOccuppyingCells((Actor*)&player, CELL_ENVIRONMENT_FLAG | ENVIRONMENT_CHASM) == 4)
        {
            // Player has fallen down a chasm and loses all health.
            playerLoseHealth(maxPlayerHealth, DAMAGE_SOURCE_CHASM);
        }
        else if (actorOccuppyingCells((Actor*)&player, CELL_ENVIRONMENT_FLAG | ENVIRONMENT_WATER) == 4)
        {
            // Player has fallen in the water without fins and loses all health.
            playerLoseHealth(maxPlayerHealth, DAMAGE_SOURCE_WATER);
        }

        // Reset the environment cells flag once we are done.
        player.environmentCells = 0;
    }

    if (levelTimerSecondsRemaining == 0)
    {
        // Time has run out so lose all health.
        playerLoseHealth(maxPlayerHealth, DAMAGE_SOURCE_TIME);
    }

    if (roomChange)
    {
        // Map needs to be redrawn on room change.
        mapRedraw = TRUE;

        // Player needs to be redrawn on room change.
        player.redraw = TRUE;

        // Stop crumbling once changing room.
        // I used to call initCrumbleQueue() here but I turned it off as 
        // it resulted in cells not crumbling as you changed rooms.
        crumblingTimerActive = FALSE;
    }
}

u8 processCell(u8 x, u8 y)
{
    return processCellInRoom(x, y, currentRoomX, currentRoomY);
}

u8 processCellInRoom(u8 x, u8 y, u8 roomX, u8 roomY)
{
    u8 cell = mapCell(x, y, roomX, roomY);
    u8 cellType = cell & CELL_GFX_RANGE;

    // Check what type of cell it is.
    if (cell & CELL_ENVIRONMENT_FLAG)
    {
        // All environment cells are non solid.
        // Different effects will occur on processing environment eventually.
        if (cellType == ENVIRONMENT_LAVA)
        {
            // Lose all player health on touching lava.
            playerLoseHealth(maxPlayerHealth, DAMAGE_SOURCE_LAVA);
        }
        else if (cellType == ENVIRONMENT_HIDDEN_TRIGGER)
        {
            // Play the switch sound effect on stepping on the hidden trigger.
            playSoundEffect(SFX_SWTICH, CHANNEL_A, 4);

            // Activate the interactable that the hidden trigger is linked to.
            activateLinkedInteractable(x, y);

            // The trigger cell can only be activated once so will be set to the default cell after use.
            updateMapCell(x, y, roomX, roomY, CLEAR_CELL);
        }
        else if (cellType == ENVIRONMENT_CHASM)
        {
            player.environmentCells = 1;
            return 0;
        }
        else if (cellType == ENVIRONMENT_WATER)
        {
            // Player can enter the water safely if they are carrying fins.
            if (!inventoryContainsItem(ITEM_FINS))
            {
                player.environmentCells = 1;
            }
            else
            {
                // Player in water which will change the shark logic.
                player.inWater = TRUE;
            }

            return 0;
        }
        else if (cellType >= ENVIRONMENT_CRUMBLING_BLOCK_5 && cellType <= ENVIRONMENT_CRUMBLING_BLOCK_1)
        {
            // Crumbling block detected so trigger the next step in the crumble cycle.
            crumbleBlock(x, y, roomX, roomY);

            return 0;
        }
        else if (cellType >= ENVIRONMENT_TURRET_LEFT && cellType <= ENVIRONMENT_TURRET_DOWN)
        {
            // Turrets are solid.
            return 1;
        }
        else if (cellType == ENVIRONMENT_HIDDEN_SOLID)
        {
            // Hidden solid cells are effectively invisible walls and are returned solid.
            return 1;
        }
        else if (cellType == ENVIRONMENT_HEDGE_0 || cellType == ENVIRONMENT_HEDGE_1 || cellType == ENVIRONMENT_HEDGE_2)
        {
            // Check if the player has a machete in their inventory.
            // Make sure it is hedge 1 or 2 to prevent the machete from affecting hedge 0.
            if (cellType > ENVIRONMENT_HEDGE_0 && inventoryContainsItem(ITEM_MACHETE))
            {
                // Check the room the machete is being used in to prevent it from activating
                // at the edge of the screen.
                if (currentRoomX == roomX && currentRoomY == roomY)
                {
                    // Start the hedge timer for this environment cell.
                    startHedgeTimer(x, y, roomX, roomY);

                    // Indicate that the machete is being actively used.
                    macheteActive_ = TRUE;
                }
            }

            // Hedges are solid.
            return 1;
        }
        else if (cellType == ENVIRONMENT_DANGEROUS_PLANT || cellType == ENVIRONMENT_DANGEROUS_PLANT_WATER)
        {
            // Running into a dangerous plant is treated like a spike.
            if (!collisionActive)
            {
                player.spikeDetected = TRUE;
            }

            return 0;
        }
        else
        {
            return 0;
        }
    }
    else if (cell & CELL_ITEM_FLAG)
    {
        // The cell used when the item is removed, if the item is a water item we use environment water cells.
        u8 updatedCell = cellType < ITEM_WATER_SHELL ? CLEAR_CELL : CELL_ENVIRONMENT_FLAG + ENVIRONMENT_WATER;

        // Check if the player is already carrying the item.
        if (inventoryContainsItem(cellType))
        {
            // Returning a 1 is equivalent to a tile (removing item flag) so that the player cannot pickup the item matching the one in the inventory.
            return 1;
        }

        // Set map cell to empty so the item is removed from the map grid.
        // Draw the cell after updating to ensure that the item is always removed after pickup.
        updateMapCellRedraw(x, y, roomX, roomY, updatedCell);

        // Add the item to the inventory for all supported items.
        addItemToInventory(cellType);

        // Play a sound effect on picking up an item.
        playSoundEffect(SFX_PICKUP, CHANNEL_A, 4);

        // Found an item so add it to the score.
        levelScore += itemScore(cellType);

        if (cellType == ITEM_HEALTH)
        {
            if (player.health + healthItemAmount <= maxPlayerHealth)
            {
                player.health += healthItemAmount;
            }
            else
            {
                player.health = maxPlayerHealth;
            }
        }
        else if (cellType == ITEM_ARROW)
        {
            player.ammo += arrowPickupCount;
        }
        else if (cellType == ITEM_GLASSES)
        {
            // Picking up the glasses will cause the trigger cells to become visible.
            // The map needs to be redrawn to show the trigger cells.
            mapRedraw = TRUE;

            // The player needs to be redrawn following the map redraw otherwise the player will become invisible.
            player.redraw = TRUE;
        }
        else if (cellType == ITEM_HOURGLASS)
        {
            // Level timer seconds cannot be more than 255.
            // Do a check to prevent an overflow before adding the time.
            if (255 - levelTimerSecondsRemaining > hourglassTime)
            {
                levelTimerSecondsRemaining += hourglassTime;
            }
            else
            {
                levelTimerSecondsRemaining = 255;
            }

            playSoundEffect(SFX_PICKUP_SPECIAL, CHANNEL_A, 4);
        }
        else if (cellType == ITEM_STAFF_BOTTOM || cellType == ITEM_STAFF_MIDDLE || cellType == ITEM_STAFF_TOP)
        {
            // Each time a staff piece is picked up the artefact pieces count is decremented.
            --remainingArtefactPieces;

            // Update the border if all the pieces have been collected.
            updateBorder();

            playSoundEffect(SFX_PICKUP_SPECIAL, CHANNEL_A, 4);
        }

        // Redraw the hud with the updated score.
        hudRedraw = TRUE;
    }
    else if (cell & CELL_INTERACTABLE_FLAG)
    {
        if (isInteractableReady(cellType))
        {
            if (cellType == INTERACTABLE_DOOR_RED || cellType == INTERACTABLE_DOOR_GREEN || cellType == INTERACTABLE_DOOR_YELLOW)
            {
                // Set map cell to empty so the door interactable is removed from the map grid.
                // Redraw the cell without the door interactable.
                // As soon as the key is used the door will disappear.
                updateMapCellRedraw(x, y, roomX, roomY, CLEAR_CELL);

                // Play the door sound effect when the door is opened.
                playSoundEffect(SFX_DOOR, CHANNEL_A, MAX_VOL);

                // Redraw the hud for any inventory changes (such as keys used).
                hudRedraw = TRUE;
            }
            else if (cellType == INTERACTABLE_PRESSURE_PLATE_2)
            {
                if (positionContainsLink(x, y))
                {
                    // I tried the switch sound effect here but instead it was a continuous beep
                    // as the player walked over it.

                    // Save the source position of the link which the player is currently on.
                    player.activatedLinkX = x;
                    player.activatedLinkY = y;
                }

                return 0;
            }
            else if (cellType == INTERACTABLE_PRESSURE_PLATE_2_BLOCK)
            {
                if (inventoryContainsItem(ITEM_BLOCK))
                {
                    // If the player is currently carrying a block then they can't go onto the pressure plate with block.
                    return 1;
                }
                else
                {
                    // Player not carrying a block so the pressure plat with block will be changed to a pressure plate
                    // and then put back into their inventory.

                    // Change the pressure plate block back to a pressure plate.
                    // Redraw the cell after the change of type.
                    updateMapCellRedraw(x, y, roomX, roomY, CELL_INTERACTABLE_FLAG + INTERACTABLE_PRESSURE_PLATE_2);

                    // Activate the linked interactable to reset its state for when the player moves of the pressure plate.
                    activateLinkedInteractable(x, y);

                    // Put the block back into the player inventory.
                    addItemToInventory(ITEM_BLOCK);

                    // Redraw hud after updating inventory.
                    hudRedraw = TRUE;
                }

                return 0;
            }
            else if (cellType == INTERACTABLE_TRAPDOOR)
            {
                // The trapdoor is activated on the player touching the interactable.
                activateInteractable(x, y, roomX, roomY, 0);
                return 0;
            }
            else if (cellType == INTERACTABLE_SPIKE_UP_LAND || cellType == INTERACTABLE_TOGGLE_SPIKE_UP || cellType == INTERACTABLE_TIMED_SPIKE_UP || cellType == INTERACTABLE_SEA_URCHIN)
            {
                // If player is invulnerable it does not matter if there is a spike.
                if (!collisionActive)
                {
                    // Set spike detected flag to later process if the player
                    // has collided with a spike.
                    player.spikeDetected = TRUE;
                }

                // Return 0 so they can fully collide with the spike.
                return 0;
            }
            else if (cellType == INTERACTABLE_SPIKE_TRAP || cellType == INTERACTABLE_SPIKE_TRAP_ACTIVATED)
            {
                // Change the spike trap to activated.
                // Redraw the cell after the change of type.
                updateMapCellRedraw(x, y, roomX, roomY, CELL_INTERACTABLE_FLAG + INTERACTABLE_SPIKE_TRAP_ACTIVATED);

                // When the player collides with the spike trap it will trigger the invulnerability timer.
                // However this only applies to practice, normal and hard.
                // If playing on hard or brutal the invulnerability timer will not be activated.
                // This makes the game a little bit easier in normal but retains the original experience
                // on hard and brutal.
                if (!collisionActive || chosenDifficulty > NORMAL_DIFFICULTY)
                {
                    // Marks the cell as a standard spike.
                    player.spikeDetected = TRUE;
                }

                // Return 0 so they can fully collide with the spike trap.
                return 0;
            }
            else if (cellType == INTERACTABLE_ONE_WAY_BLOCK_RIGHT || cellType == INTERACTABLE_SWITCHABLE_ONE_WAY_BLOCK_RIGHT)
            {
                // When the player is moving left (negative x velocity) it will be solid.
                // If the player is moving up, down or right it will be non solid.
                // Using the not to allow up and down so the player cannot be stuck in a block.
                if (!(player.object.velocity.x < 0))
                {
                    return 0;
                }
            }
            else if (cellType == INTERACTABLE_ONE_WAY_BLOCK_LEFT || cellType == INTERACTABLE_SWITCHABLE_ONE_WAY_BLOCK_LEFT)
            {
                // When the player is moving right (positive x velocity) it will be solid.
                // If the player is moving up, down or left it will be non solid.
                // Using the not to allow up and down so the player cannot be stuck in a block.
                if (!(player.object.velocity.x > 0))
                {
                    return 0;
                }
            }
            else if (cellType == INTERACTABLE_ONE_WAY_BLOCK_UP || cellType == INTERACTABLE_SWITCHABLE_ONE_WAY_BLOCK_UP)
            {
                // When the player is moving down (positive y velocity) it will be solid.
                // If the player is moving up, left or right it will be non solid.
                // Using the not to allow left and right so the player cannot be stuck in a block.
                if (!(player.object.velocity.y > 0))
                {
                    return 0;
                }
            }
            else if (cellType == INTERACTABLE_ONE_WAY_BLOCK_DOWN || cellType == INTERACTABLE_SWITCHABLE_ONE_WAY_BLOCK_DOWN)
            {
                // When the player is moving up (negative y velocity) it will be solid.
                // If the player is moving down, left or right it will be non solid.
                // Using the not to allow left and right so the player cannot be stuck in a block.
                if (!(player.object.velocity.y < 0))
                {
                    return 0;
                }
            }
            else
            {
                // Everything but INTERACTABLE_GATE is not solid but gate will never get here as isInteractableReady will always return FALSE for a gate.
                // Return 0 for INTERACTABLE_CHECKPOINT, INTERACTABLE_CHECKPOINT_ACTIVE, INTERACTABLE_TOGGLE_SPIKE_DOWN, INTERACTABLE_TIMED_SPIKE_DOWN,
                // INTERACTABLE_LADDER, INTERACTABLE_PHASE_BLOCK_OFF, INTERACTABLE_PHASE_BLOCK_ON, INTERACTABLE_LEVER_ON, INTERACTABLE_LEVER_OFF,
                // and INTERACTABLE_EXIT.
                return 0;
            }
        }
    }

    return cell;
}

void moveRoomBelow()
{
    // Update player position for the new room.
    player.object.position.y = 0;
    player.object.previousPosition.y = 0;

    // Update the current room.
    currentRoom += 3;
    ++currentRoomY;
    player.object.position.room = currentRoom;
    player.object.previousPosition.room = currentRoom;
    roomChange = TRUE;
}

void moveRoomAbove()
{
    // Update player position for the new room above.
    // Using -24 instead of -16 prevents the hud being drawn over by the player sprite.
    player.object.position.y = SCREEN_Y - 24;
    player.object.previousPosition.y = SCREEN_Y - 24;

    // Update the current room.
    currentRoom -= 3;
    --currentRoomY;
    player.object.position.room = currentRoom;
    player.object.previousPosition.room = currentRoom;
    roomChange = TRUE;
}

void moveRoomX(u8 newX, i8 roomDeltaX)
{
    // Update player position for the new room.
    player.object.position.x = newX;
    player.object.previousPosition.x = newX;
    player.object.velocity.x = 0;

    // Update the current room.
    currentRoom += roomDeltaX;
    currentRoomX += roomDeltaX;
    player.object.position.room = currentRoom;
    player.object.previousPosition.room = currentRoom;
    roomChange = TRUE;
}

void processSpikeCollision()
{
    u8 spikeCollision = 0;

    u8 occupiedXTileLeft = (player.object.position.x) >> 2;
    u8 occupiedXTileRight = (player.object.position.x + TILE_WIDTH_BYTES - 1);
    u8 occupiedYTileBottom = (player.object.position.y + TILE_HEIGHT - 1);
    u8 occupiedYTileTop = (player.object.position.y) >> 4;
    occupiedXTileRight >>= 2;
    occupiedYTileBottom >>= 4;

    spikeCollision += cellContainsSpike(occupiedXTileLeft, occupiedYTileTop);
    spikeCollision += cellContainsSpike(occupiedXTileRight, occupiedYTileTop);
    spikeCollision += cellContainsSpike(occupiedXTileLeft, occupiedYTileBottom);
    spikeCollision += cellContainsSpike(occupiedXTileRight, occupiedYTileBottom);

    if (spikeCollision)
    {
        // Different spike deaths for in and out of water.
        if (player.inWater)
        {
            playerLoseHealth(spikeDamageAmount, DAMAGE_SOURCE_SEA_URCHIN);
        }
        else
        {
            playerLoseHealth(spikeDamageAmount, DAMAGE_SOURCE_SPIKE);
        }

        // Invulnerability timer.
        activateCollisionTimer();

        // Update health on the hud.
        hudRedraw = TRUE;
    }

    // Reset the spike detected flag after the spike detection has been processed.
    player.spikeDetected = FALSE;
}

u8 cellContainsSpike(u8 x, u8 y)
{
    // Retrieve the cell in the current room.
    u8 cell = mapCell(x, y, currentRoomX, currentRoomY);

    // Check if the cell is a spike.
    if (cell == CELL_INTERACTABLE_FLAG + INTERACTABLE_SPIKE_UP_LAND ||
        cell == CELL_INTERACTABLE_FLAG + INTERACTABLE_SPIKE_TRAP_ACTIVATED ||
        cell == CELL_INTERACTABLE_FLAG + INTERACTABLE_TOGGLE_SPIKE_UP ||
        cell == CELL_INTERACTABLE_FLAG + INTERACTABLE_TIMED_SPIKE_UP ||
        cell == CELL_INTERACTABLE_FLAG + INTERACTABLE_SEA_URCHIN ||
        cell == CELL_ENVIRONMENT_FLAG + ENVIRONMENT_DANGEROUS_PLANT ||
        cell == CELL_ENVIRONMENT_FLAG + ENVIRONMENT_DANGEROUS_PLANT_WATER)
    {
        return TRUE;
    }

    return FALSE;
}

void playerLoseHealth(u8 amount, u8 damageSource)
{
    if (player.health - amount > 0)
    {
        player.health -= amount;

        // Play damage sound if the player takes damage from anything.
        playSoundEffect(SFX_TAKE_DAMAGE, CHANNEL_A, 4);
    }
    else
    {
        if (playerLives--)
        {
            levelState = LEVEL_DEATH;
            gameRunning = FALSE;
            player.health = playerStartingHealth;

            // Player is dead so display death graphic.
            playerDeath(damageSource);
        }
    }
}

void spawnProjectile()
{
    // Not on top of an interactable so shoot instead (use and shoot are shared).
    // Only launch a projectile if one has not previously been launched.
    if (!projectileState[PLAYER_PROJECTILE_INDEX].health)
    {
        // Ensure the player has a weapon with enough ammo.
        if (inventoryContainsItem(ITEM_BOW) && player.ammo)
        {
            // Initialise the projectile to the player's current position.
            initProjectile(&projectileState[PLAYER_PROJECTILE_INDEX], player.object.position.x, player.object.position.y, player.direction);

            // Save the player actor id to the projectile for identifying that the projectile was spawned by the player.
            projectileState[PLAYER_PROJECTILE_INDEX].actorId = player.id;

            // Save the projectile queue index to the projectile state.
            projectileState[PLAYER_PROJECTILE_INDEX].projectileQueueIndex = PLAYER_PROJECTILE_INDEX;

            // Reduce the player ammo after firing.
            --player.ammo;

            // Play a sound when the projectile is created.
            playSoundEffect(SFX_SHOOT, CHANNEL_A, 4);

            // Redraw the hud to display the updated ammo value.
            hudRedraw = TRUE;
        }
    }
}

void teleportPlayer(u8 destinationX, u8 destinationY, u8 destinationRoomX, u8 destinationRoomY)
{
    // Update player position.
    player.object.position.x = destinationX * TILE_WIDTH_BYTES;
    player.object.position.y = destinationY * TILE_HEIGHT;
    player.object.previousPosition.x = destinationX * TILE_WIDTH_BYTES;
    player.object.previousPosition.y = destinationY * TILE_HEIGHT;

    // Reset velocity.
    player.object.velocity.x = 0;
    player.object.velocity.y = 0;

    // Update the tracking for the current room.
    currentRoom = destinationRoomX + destinationRoomY * 3;
    currentRoomX = destinationRoomX;
    currentRoomY = destinationRoomY;

    // Update player room.
    player.object.position.room = currentRoom;
    player.object.previousPosition.room = currentRoom;

    // Player sound effect on teleporting the player (phase block, ladder or trapdoor)
    playSoundEffect(SFX_TELEPORT, CHANNEL_A, 6);

    // Indicate the room has changed.
    roomChange = TRUE;
}

void playerDeath(u8 damageSource)
{
    // Using the collision timer instead of a death timer.
    collisionActive = TRUE;
    collisionTimerCount = 0;

    // Updates the player to the appropriate death graphic.
    player.graphicIndex = damageSource + PLAYER_GRAPHIC_DEAD_OFFSET;

    // Make sure to enable redraw before drawing the player.
    player.redraw = TRUE;

    while (collisionActive)
    {
        // If we draw the player before the loop it compiles in 
        drawPlayer();
    }
}

void playerEnemyCollision()
{
    u8 i = 0;
    Enemy* enemy = mapEnemies;

    // If player is invulnerable it does not matter if there is a collision.
    if (!collisionActive)
    {
        for (i = 0; i < mapEnemyCount; ++i)
        {
            // Check if the enemy is on screen.
            if (enemy->object.position.room == currentRoom)
            {
                // Timed spike handles collision using interactable rather than enemy object.
                if (enemy->type != ENEMY_TIMED_SPIKE)
                {
                    // Check for overlap.
                    if (difference(enemy->object.position.y, player.object.position.y) < TILE_HEIGHT && 
                        difference(enemy->object.position.x, player.object.position.x) < TILE_WIDTH_BYTES)
                    {
                        if (enemy->type >= ENEMY_MOVING_ROCK_LEFT && enemy->type <= ENEMY_MOVING_ROCK_DOWN)
                        {
                            // Player gets crushed by moving block causing instant death.
                            playerLoseHealth(maxPlayerHealth, DAMAGE_SOURCE_CRUSHED);
                        }
                        else
                        {
                            // Activate collision timer to make player temporarily invulnerable.
                            activateCollisionTimer();

                            if (enemy->type == ENEMY_SHARK)
                            {
                                // Player loses health once they intersect with an enemy.
                                playerLoseHealth(spikeDamageAmount, DAMAGE_SOURCE_SHARK);
                            }
                            else
                            {
                                // Player loses health once they intersect with an enemy.
                                playerLoseHealth(spikeDamageAmount, DAMAGE_SOURCE_ENEMY);
                            }
                        }

                        // Redraw the hud with the new player health amount.
                        hudRedraw = TRUE;

                        // No need to check for more collisions.
                        break;
                    }
                }
            }

            // Move to the next enemy in the array.
            ++enemy;
        }
    }
}

void dropBlockItem(u8 cellX, u8 cellY)
{
    // It is assumed that the cellX and cellY passed to the function is one square from the player's current position (x,y).
    // Make sure that the cell values are not outside the current room.
    if (cellX < SCREEN_TILES_X && cellY < SCREEN_TILES_Y)
    {
        u8 cell = mapCell(cellX, cellY, currentRoomX, currentRoomY);

        // Check that the position we want to drop the block on is empty.
        // I don't want it changing anything once it is removed so it can't be dropped on anything after crack.
        if (cell & CELL_ENVIRONMENT_FLAG)
        {
            if (cell <= CELL_ENVIRONMENT_FLAG + ENVIRONMENT_FLOWERS)
            {
                // Update the empty cell to the item block.
                // Redraw the cell with the item block in it.
                updateMapCellRedraw(cellX, cellY, currentRoomX, currentRoomY, CELL_ITEM_FLAG | ITEM_BLOCK);

                // Remove the block from the player inventory.
                inventoryItems &= ~INVENTORY_ITEM_BLOCK;

                // Redraw the hud to update the inventory on the hud.
                hudRedraw = TRUE;
            }
            else if (cell == CELL_ENVIRONMENT_FLAG + ENVIRONMENT_CHASM)
            {
                // Update the chasm cell to the default environment cell.
                // Redraw the cell following filling the chasm cell.
                updateMapCellRedraw(cellX, cellY, currentRoomX, currentRoomY, CLEAR_CELL);

                // Remove the block from the player inventory.
                inventoryItems &= ~INVENTORY_ITEM_BLOCK;

                // Redraw the hud to update the inventory on the hud.
                hudRedraw = TRUE;
            }
        }
        else if (cell == CELL_INTERACTABLE_FLAG + INTERACTABLE_PRESSURE_PLATE_2)
        {
            // Change pressure plate to pressure plate with block.
            // Redraw the cell following changing the pressure plate to the pressure plate with block.
            updateMapCellRedraw(cellX, cellY, currentRoomX, currentRoomY, CELL_INTERACTABLE_FLAG | INTERACTABLE_PRESSURE_PLATE_2_BLOCK);

            // Activates the pressure plate with block.
            activateInteractable(cellX, cellY, currentRoomX, currentRoomY, 0);

            // Remove the block from the player inventory.
            inventoryItems &= ~INVENTORY_ITEM_BLOCK;

            // Redraw the hud after updating inventory.
            hudRedraw = TRUE;
        }
    }
}