#ifndef PLAYER_H
#define PLAYER_H

#include <cpctelera.h>
#include "Actor.h"
#include "Environment.h"
#include "GameTypes.h"
#include "Hud.h"
#include "Interactable.h"
#include "Interrupt.h"
#include "Inventory.h"
#include "Item.h"
#include "Keys.h"
#include "Map.h"
#include "Object.h"
#include "Projectile.h"
#include "ProjectileState.h"
#include "sprites/PlayerSprites.h"

// The death of the player is from running out of time.
#define DAMAGE_SOURCE_TIME 0

// The damage or death of the player is from a spike.
#define DAMAGE_SOURCE_SPIKE 1

// The death of the player is from a chasm.
#define DAMAGE_SOURCE_CHASM 2

// The death of the player is from lava.
#define DAMAGE_SOURCE_LAVA 3

// The death of the player is from being crushed by a rock.
#define DAMAGE_SOURCE_CRUSHED 4

// The death of the player is from drowning in water (no fins).
#define DAMAGE_SOURCE_WATER 5

// The death of the player is from being shot by a projectile.
#define DAMAGE_SOURCE_PROJECTILE 6

// The death of the player is from being eaten by an enemy shark.
#define DAMAGE_SOURCE_SHARK 7

// The death of the player is from a land enemy.
#define DAMAGE_SOURCE_ENEMY 8

// The death of the player is from a sea urchin.
#define DAMAGE_SOURCE_SEA_URCHIN 9

#define PLAYER_GRAPHIC_RIGHT 0
#define PLAYER_GRAPHIC_LEFT 1
#define PLAYER_GRAPHIC_DOWN 2
#define PLAYER_GRAPHIC_UP 3
#define PLAYER_GRAPHIC_RIGHT_WATER 4
#define PLAYER_GRAPHIC_LEFT_WATER 5
#define PLAYER_GRAPHIC_DOWN_WATER 6
#define PLAYER_GRAPHIC_UP_WATER 7
#define PLAYER_GRAPHIC_DEAD_TIME 8
#define PLAYER_GRAPHIC_DEAD_IMPALED 9
#define PLAYER_GRAPHIC_DEAD_CHASM 10
#define PLAYER_GRAPHIC_DEAD_LAVA 11
#define PLAYER_GRAPHIC_DEAD_CRUSHED 12
#define PLAYER_GRAPHIC_DEAD_WATER 13
#define PLAYER_GRAPHIC_DEAD_SHOT 14
#define PLAYER_GRAPHIC_DEAD_SHARK 15
#define PLAYER_GRAPHIC_DEAD_ENEMY 16
#define PLAYER_GRAPHIC_DEAD_IMPALED_WATER 17

// Offset used as the start of the player dead graphics in the player graphic array.
#define PLAYER_GRAPHIC_DEAD_OFFSET PLAYER_GRAPHIC_DEAD_TIME

typedef struct Player Player;

struct Player
{
    // Actor object containing position and velocity.
    Object object;

    // The x position of the link activated by the player.
    u8 activatedLinkX;

    // The y position of the link activated by the player.
    u8 activatedLinkY;

    // Number of hitpoints.
    u8 health;

    // Unique identifier for identifying the different actors in the system.
    u8 id;

    // Indicates whether the actor should be redrawn on screen.
    u8 redraw;

    // The type of actor (player, enemy, block).
    u8 actorType;

    // Flag to indicate if a spike has been detected when processing a cell.
    u8 spikeDetected;

    // Detecting one of four environment cells that the player could be in.
    u8 environmentCells;

    // The direction that the player is currently facing.
    u8 direction;

    // The previous direction that the player was facing.
    u8 previousDirection;

    // The amount of ammo the player is carrying.
    u8 ammo;

    // Indicates that the player is in the water that is used by sharks.
    u8 inWater;

    // The index of the graphic representing the player in game.
    u8 graphicIndex;
};

// Global player object.
extern Player player;

// Initialises the player struct with initial values.
void initPlayer(u8 x, u8 y, u8 loadingCheckpoint);

// Draws the player on the screen as positioned in the player object.
void drawPlayer();

// Updates the player velocity based on user input.
void movePlayer();

// Player logic such as interactions and velocity/position.
void logicPlayer();

// Processes the cell in the current room that the player is potentially occupying.
u8 processCell(u8 x, u8 y);

// Processes the cell in the specified room that the player is potentially occupying.
u8 processCellInRoom(u8 x, u8 y, u8 roomX, u8 roomY);

// Moves the player to the room below the current one.
void moveRoomBelow();

// Moves the player to the room above the current one.
void moveRoomAbove();

// Moves the screen to a room on the x axis (left or right).
void moveRoomX(u8 newX, i8 roomDeltaX);

// Tests whether the player has collided with a spike (as detected) and updates the player health if there has been a collision.
void processSpikeCollision();

// Indicates whether the specified cell in the current room contains a spike.
u8 cellContainsSpike(u8 x, u8 y);

// Player has been damaged and will lose health and possible lives as a result.
void playerLoseHealth(u8 amount, u8 damageSource);

// Creates a projectile when the player has a weapon with ammo.
void spawnProjectile();

// Moves the player to a new location on the map.
void teleportPlayer(u8 destinationX, u8 destinationY, u8 destinationRoomX, u8 destinationRoomY);

// Displays the player death graphic based on the damage source.
void playerDeath(u8 damageSource);

// Detect whether the player has collided with enemies on screen.
void playerEnemyCollision();

// Drops the block the player is currently carrying and drops it in front of the player.
void dropBlockItem(u8 cellX, u8 cellY);

#endif