#include "Projectile.h"

#define PROJECTILE_WIDTH_ 1
#define PROJECTILE_HEIGHT_ 4

// Bright yellow projectile.
#define PROJECTILE_YELLOW_ 1

// Bright red projectile.
#define PROJECTILE_RED_ 3

// The colour of the projectile drawn on the screen.
u8 projectileColour_;

void initProjectile(Projectile* projectile, u8 x, u8 y, u8 direction)
{
    projectile->redraw = TRUE;
    projectile->direction = direction;
    projectile->health = 1;
    projectile->object.velocity.x = 0;
    projectile->object.velocity.y = 0;
    projectile->horizontalMovementFlag = 0;
    projectile->object.position.room = currentRoom;

    // Initially the projectile is not in the projectile queue so 255 is used as unset.
    projectile->projectileQueueIndex = 255;

    // Spawning the projectiles with space prevents the projectile from copying the background where the actor is.
    if (direction == DIRECTION_RIGHT)
    {
        projectile->object.position.x = x + 6; // Tile width (4), projectile width (1), and a space (1).
        projectile->object.position.y = y + 6; // Half the tile height (8 of 16) minus half the projectile height (2 of 4).

        projectile->object.velocity.x = 1;
    }
    else if (direction == DIRECTION_LEFT)
    {
        projectile->object.position.x = x - 3; // Projectile width (1) and a space (2).
        projectile->object.position.y = y + 6; // Half the tile height (8 of 16) minus half the projectile height (2 of 4).

        projectile->object.velocity.x = -1;
    }
    else if (direction == DIRECTION_UP)
    {
        projectile->object.position.x = x + 2; // Half Tile width (2 of 4).
        projectile->object.position.y = y - 8; // Projectile height (4) and a space (4).

        projectile->object.velocity.y = -1;
    }
    else if (direction == DIRECTION_DOWN)
    {
        projectile->object.position.x = x + 2; // Half Tile width (2 of 4).
        projectile->object.position.y = y + 21; // Tile height (16), projectile height (4) and a space (1).

        projectile->object.velocity.y = 1;
    }

    projectile->object.previousPosition.x = projectile->object.position.x;
    projectile->object.previousPosition.y = projectile->object.previousPosition.y;

    // Copy the content of the background where the projectile will be drawn to save for later.
    copyBackground(projectile);
}

void drawProjectile(Projectile* projectile)
{
    // Do not draw the projectile if health is zero (it has been destroyed).
    if (projectile->health && projectile->redraw)
    {
        // Draw over the position the projectile previously was with the saved background sprite.
        u8* memory = cpct_getScreenPtr(CPCT_VMEM_START, projectile->object.previousPosition.x, projectile->object.previousPosition.y);
        cpct_drawSprite(projectile->backgroundSprite, memory, PROJECTILE_WIDTH_, PROJECTILE_HEIGHT_);

        // Copy the content of the background where the projectile sprite currently is (before it is drawn) for later redraw when the projectile moves.
        copyBackground(projectile);

        // Draw a square to represent the projectile at the current position.
        memory = cpct_getScreenPtr(CPCT_VMEM_START, projectile->object.position.x, projectile->object.position.y);

        cpct_drawSolidBox(memory, cpctm_px2byteM0(projectileColour_, projectileColour_), PROJECTILE_WIDTH_, PROJECTILE_HEIGHT_);

        projectile->redraw = FALSE;
    }
}

void copyBackground(Projectile* projectile)
{
    u8* memory = cpct_getScreenPtr(CPCT_VMEM_START, projectile->object.position.x, projectile->object.position.y);
    cpct_getScreenToSprite(memory, projectile->backgroundSprite, PROJECTILE_WIDTH_, PROJECTILE_HEIGHT_);
}

void logicProjectile(Projectile* projectile)
{
    // Do not update the projectile if health is zero (it has been destroyed).
    if (projectile->health)
    {
        u8 cellX;
        u8 cellY;
        u8 cell;

        // Projectile defaults to bright red (or green equivalent).
        projectileColour_ = PROJECTILE_RED_;

        // Update the previous position before updating the current position with the velocity.
        projectile->object.previousPosition.x = projectile->object.position.x;
        projectile->object.previousPosition.y = projectile->object.position.y;

        // Calculate new projectile position from velocity.
        projectile->object.position.y += projectile->object.velocity.y;

        // Only update the projectiles horizontal position every second iteration.
        // This ensures that the projectile appears to move at the same velocity virtually as horizontally.
        // Also it ensure the projectile moves at the same speed as the player.
        if (projectile->horizontalMovementFlag++)
        {
            projectile->object.position.x += projectile->object.velocity.x;

            // Reset flag to ensure projectile can only move horizontally every second iteration.
            projectile->horizontalMovementFlag = 0;
        }

        // The cells that the projectile is occupying.
        cellX = projectile->object.position.x / TILE_WIDTH_BYTES;
        cellY = projectile->object.position.y / TILE_HEIGHT;
        cell = mapCell(cellX, cellY, currentRoomX, currentRoomY);

        projectile->redraw = TRUE;

        // Activate the interactable lever if the projectile hit its.
        if (cell == CELL_INTERACTABLE_FLAG + INTERACTABLE_LEVER_ON || cell == CELL_INTERACTABLE_FLAG + INTERACTABLE_LEVER_OFF)
        {
            // Activate the lever to switch between on and off.
            activateInteractable(cellX, cellY, currentRoomX, currentRoomY, 0);

            // Redraw the lever after it has been activated.
            drawCellAtPosition(cellX, cellY);
        }

        if (cell > CELL_EMPTY && cell < CELL_ENVIRONMENT_FLAG)
        {
            // Destroy the projectile if it runs into a tile, item or interactable (environment is not solid).
            destroyProjectile(projectile);

            // Redraw the item block otherwise we can end up with a block with a projectile size hole in it.
            if (cell == CELL_ITEM_FLAG + ITEM_BLOCK)
            {
                drawCellAtPosition(cellX, cellY);
            }
        }
        else if (cell == CELL_ENVIRONMENT_FLAG + ENVIRONMENT_DANGEROUS_PLANT || cell == CELL_ENVIRONMENT_FLAG + ENVIRONMENT_DANGEROUS_PLANT_WATER)
        {
            destroyProjectile(projectile);

            // Updates the cell for the dangerous plant based on whether it is a water plant or not.
            if (cell == CELL_ENVIRONMENT_FLAG + ENVIRONMENT_DANGEROUS_PLANT)
            {
                updateMapCellRedraw(cellX, cellY, currentRoomX, currentRoomY, CLEAR_CELL);
            }
            else
            {
                updateMapCellRedraw(cellX, cellY, currentRoomX, currentRoomY, CELL_ENVIRONMENT_FLAG + ENVIRONMENT_WATER);
            }

            // Play a sound when the plant is destroyed.
            playSoundEffect(SFX_ENEMY_TAKE_DAMAGE, CHANNEL_A, 4);

            // Add 100 points for destroying a plant.
            levelScore += 10;

            // Redraw the hud to show score update.
            hudRedraw = TRUE;
        }
        else if (cell >= CELL_ENVIRONMENT_FLAG + ENVIRONMENT_HEDGE_0)
        {
            // Destroy the projectile when shooting hedge or hidden solid
            destroyProjectile(projectile);
        }
        else if (cellX >= SCREEN_TILES_X)
        {
            // Destroy the tile if it goes off screen.
            destroyProjectile(projectile);
        }
        else if (cellY >= SCREEN_TILES_Y)
        {
            // Destroy the tile if it goes off screen.
            destroyProjectile(projectile);
        }
        else if (projectile->object.position.room != currentRoom)
        {
            // Destroy projectile if it goes out of the current room.
            destroyProjectile(projectile);
        }
        else if (cell == CELL_ENVIRONMENT_FLAG + ENVIRONMENT_LAVA)
        {
            // Change the colour of the projectile from red to yellow when it is over lava to make it visible.
            projectileColour_ = PROJECTILE_YELLOW_;
        }

        // If projectiles interact with actors then the actor is damaged and the projectile is destroyed.
        actorProjectileColliction(projectile);

        // If projectiles intersect then both are destroyed.
        processProjectileIntersection(projectile);
    }
}

void destroyProjectile(Projectile* projectile)
{
    u8* memory = cpct_getScreenPtr(CPCT_VMEM_START, projectile->object.previousPosition.x, projectile->object.previousPosition.y);

    // The projectile queue contains the index of the enemy in the enemy array that spawned the projectile.
    u8 enemyIndex = projectileQueue[projectile->projectileQueueIndex];

    projectile->health = 0;
    projectile->redraw = FALSE;

    // Draw the projectile where the projectile was previously before it was destroyed.
    cpct_drawSprite(projectile->backgroundSprite, memory, PROJECTILE_WIDTH_, PROJECTILE_HEIGHT_);

    // Clear the projectile from the projectile queue on destruction.
    projectileQueue[projectile->projectileQueueIndex] = PROJECTILE_QUEUE_FREE;

    // Only clear the projectile from the enemy if it wasn't spawned by the player.
    if (projectile->projectileQueueIndex != PLAYER_PROJECTILE_INDEX)
    {
        // Clear the projectile from the enemy by resetting the index to 0.
        mapEnemies[enemyIndex].projectileIndex = 0;
    }
}

void processProjectileIntersection(Projectile* projectile)
{
    u8 i = 0;

    Projectile* otherProjectile = projectileState;

    for (i = 0; i < MAX_PROJECTILE_COUNT; ++i)
    {
        // Make sure it is a different projectile and that it is active.
        if (otherProjectile->health && otherProjectile->actorId != projectile->actorId)
        {
            // Check if projectiles are intersecting (width is 1 so x can be equality test).
            if (projectile->object.position.x == otherProjectile->object.position.x &&
                difference(projectile->object.position.y, otherProjectile->object.position.y) < PROJECTILE_HEIGHT_)
            {
                // Destroy both projectiles and end.
                // It is not likely that there would be more than 2 overlapping.
                destroyProjectile(projectile);
                destroyProjectile(otherProjectile);
                break;
            }
        }

        // Move to the next projectile.
        ++otherProjectile;
    }
}

void actorProjectileColliction(Projectile* projectile)
{
    Enemy* enemy = mapEnemies;
    u8 i = 0;

    for (i = 0; i < mapEnemyCount; ++i)
    {
        damageActor((Actor*)enemy, projectile);

        ++enemy;
    }

    damageActor((Actor*)&player, projectile);
}

void damageActor(_Actor* actor, Projectile* projectile)
{
    // Test if the actor (make sure its in the same room) is colliding with the projectile (based on current position).
    if (actor->object.position.room == currentRoom && testActorProjectileIntersection(actor, projectile))
    {
        // Make sure that the actor being damaged is not the one that spawned the projectile.
        if (projectile->actorId != actor->id)
        {
            // Destroy the projectile when the projectile collides with an actor.
            destroyProjectile(projectile);

            if (actor->actorType == ACTOR_TYPE_ENEMY)
            {
                // Enemy loses health on shooting the enemy.
                --actor->health;

                // Play a sound when the enemy takes damage.
                playSoundEffect(SFX_ENEMY_TAKE_DAMAGE, CHANNEL_A, 4);

                if (!actor->health)
                {
                    destroyEnemy((Enemy*)actor);

                    // Add 300 points for destroying an enemy.
                    levelScore += 30;

                    // Redraw the hud after gaining score.
                    hudRedraw = TRUE;
                }
            }
            else if (actor->actorType == ACTOR_TYPE_PLAYER)
            {
                playerLoseHealth(spikeDamageAmount, DAMAGE_SOURCE_PROJECTILE);

                // Redraw the hud after losing health.
                hudRedraw = TRUE;
            }
        }
    }
}

u8 testActorProjectileIntersection(_Actor* actor, Projectile* projectile)
{
    if (actor->object.position.x + TILE_WIDTH_BYTES > projectile->object.position.x &&
        actor->object.position.x < projectile->object.position.x + PROJECTILE_WIDTH_)
    {
        if (actor->object.position.y + TILE_HEIGHT > projectile->object.position.y &&
            actor->object.position.y < projectile->object.position.y + PROJECTILE_HEIGHT_)
        {
            return TRUE;
        }
    }

    return FALSE;
}

void initProjectiles()
{
    u8 i = 0;

    Projectile* projectile = projectileState;

    for (i = 0; i < MAX_PROJECTILE_COUNT; ++i)
    {
        // Once health is 0 it will not draw the projectile.
        projectile->health = 0;

        // Clear the projectile queue when initialising projectiles.
        projectileQueue[i] = PROJECTILE_QUEUE_FREE;

        // Move to the next projectile in the projectile state.
        ++projectile;
    }
}